;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; program:   Win3ds36 - *.3ds / *.asc files viewer + demonstration of
;;	       several graphics effects + tiny based on Bezier curves and
;;	      patches 3d objects editor + some manipulation on 3d mesh
;;	      routines.
;; compiler : flat assembler
;; author:    Maciej Guba aka macgub
;; web	 :    http://macgub.co.pl - if this site not works, search the web
;;	      for keyword 'macgub'.
;; email :    macgub@macgub.co.pl
;; This application don't uses any particular 3d graphic library.
;; Written from scratch in pure assembler. Thanks to all, who was
;; helping me to do it. Especially:
;; Jan Pawel II, Tomasz Grysztar, Madis Kalme, Mikolaj Feliks, Lostcauz,
;; Brian Paul, Reverend, Pablo Reda, MHajduk, Ica, James Foley,
;; Andries van Dam, Steven Feiner, John Hughes, Richard Phillips,
;; tthsqe, J. Burkardt, Morgan McGuire, Pierre Bezier, Przemyslaw Kiciak
;; and many others...
;;
;; Program needs SSE4 extension, in full mode.
;; When only SEE3 is present some functions are disabled.
;; To achive SEE2 version uncomment line 52.
;; See readme.txt file to read some user interface description.
;;
;; Coding style is terribly ugly. Other words it's my own mess. ;)
;;
;;	    Generaxl note about implemented operations. - Some of it may run teribly
;;	    slow aspecially on 'large' objects. Maybye some tests for low detail objects
;;	    may prevent your speed disapointing. Note that even operations on small
;;	    objects may be slow, bacause way of sloving problems.
;;	    Its 32 bit app, so object above ~10 000 000 vertices/faces may be to big
;;	    for 4GB RAM limit - some operations alloc temporaly many memory (because
;;	    using mem greedy pivot lists). Many operation cannot be launched  parallel
;;	    in the same time (ex. editing).
;;
;;
;;	    After run program shows dialog to open file.
;;	    Choose file in 3ds or asc format and ...admire. ;)
;;
;;
;;
;;				      "No work is ever complete,
;;				       and this one is no exception."
;;							Pierre Bezier
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

format PE GUI 4.0
entry start
SSE4 = 0
SSE3 = 1
Ext = SSE3   ; I switch off some sse4 lines of code
;include 'sse2.inc'	      ; uncomment to achive sse2 version

include './include/win32a.inc'
include 'macr.inc'

BXRES	   = 400    ; choose window size of next edit area
BYRES	   = 400
MXRES	   = 400    ; menu window size
MYRES	   = 450

TEX_X	   = 512    ; texture size
TEX_Y	   = 512
TEXTURE_SIZE	= (TEX_X * TEX_Y)-1
TEX_SHIFT	= 9
NEXT_CURV_NUMB	= 80   ; must be divisable by 4
		       ; next cuve allow bend operation on chunk
FONT_FIX	= 170  ; decreased  size of menu
ROUND		equ 8

PROMPTX = 500
PROMPTY = 30
INFOX	= MXRES
INFOY	= 140
WM_FINISH = WM_USER+100h

IN_CNST = INFOY  + BYRES ;  + MYRES


; some intial macros - just block of instructions..

macro stretch_menu
{
    pushad
    movzx    eax,[xres_var]
    mov      ebx,menu_screen
    mov      ecx,MYRES + IN_CNST
    movzx    edx,[yres_var]
    cmp      ecx,edx
    cmova    ecx,edx
    invoke   StretchDIBits,[hdc],eax,0,MXRES,ecx,0,0,MXRES,ecx,ebx,Mbmi,0,SRCCOPY
    popad
}

macro clrscr
{
     pushad
     movzx     eax,[whole_xres]
     movzx     ebx,[whole_yres]
     invoke    StretchDIBits,[hdc],0,0,eax,ebx,0,0,2,2,fakebitmap,Wbmi,0,SRCCOPY
     popad


}

macro do_size_vars
{
;     eax - x
;     edx - y

     and     eax,0xfffffffc
     mov     [whole_xres],ax
     sub     eax,MXRES ;+MXPRES+10    ; place for menu and bezier curves area
     and     eax,-4	  ; motion blur need that
     mov     ebx,eax
     mov     [xres_var],ax
     shr     eax,1
     mov     [xxadd],eax
  ;   mov     [tex_x_div2],eax
  ;   mov     [tex_y_div2],eax


     and     edx,0xfffffffc
     mov     [whole_yres],dx
     sub     edx,70
     and     edx,-4
     mov     [yres_var],dx
     mov     ecx,edx
     shr     ecx,1
     mov     [yyadd],ecx


     movzx    eax,[xres_var]
     movzx    ebx,[yres_var]
     sub      eax,6
     sub      ebx,6

     mov      edi,max_work_x
     mov      [edi],ax
     mov      [edi+2],bx

     dec      eax
     dec      ebx
     mov      edi,tri_area_x2
     mov      [edi],eax
     mov      [edi+4],ebx
     cvtpi2ps xmm0,[edi]
     movlps   [edi],xmm0
;     cvtsi2ss xmm0,eax
;     cvtsi2ss xmm1,ebx
;     movss    [tri_area_x2],xmm0  ; init teslate area
;     movss    [tri_area_y2],xmm1
}



macro alloc_screen_mem
{
     movzx   eax,[xres_var]
     movzx   ebx,[yres_var]
     imul    ebx,eax
     shr     ebx,2 ; 2 bit of each position (pixel)
     add     ebx,200
     invoke  VirtualAlloc,NULL,ebx,MEM_COMMIT+MEM_RESERVE,PAGE_EXECUTE_READWRITE
     mov     [edit_tri_area_ptr],eax  ; teslate area

     add     ebx,65536
     shl     ebx,2
     invoke  VirtualAlloc,NULL,ebx,MEM_COMMIT+MEM_RESERVE,PAGE_EXECUTE_READWRITE
     mov     [slices_counter_ptr],eax

     shl     ebx,2
     invoke  VirtualAlloc,NULL,ebx,MEM_COMMIT+MEM_RESERVE,PAGE_EXECUTE_READWRITE
     mov     [screen_ptr],eax

     invoke  VirtualAlloc,NULL,ebx,MEM_COMMIT+MEM_RESERVE,PAGE_EXECUTE_READWRITE
     mov     [slices_ptrs_buff_ptr],eax

     shl     ebx,2   ; in Z buffer , also 4 slices buffer
     add     ebx,65535
     invoke  VirtualAlloc,NULL,ebx,MEM_COMMIT+MEM_RESERVE,PAGE_EXECUTE_READWRITE
     mov     [Zbuffer_ptr],eax

     shr     ebx,2
     invoke  VirtualAlloc,NULL,ebx,MEM_COMMIT+MEM_RESERVE,PAGE_EXECUTE_READWRITE
     mov     [edit_buf_ptr],eax

     shr     ebx,2   ; maybye on bit ??
     invoke  VirtualAlloc,NULL,ebx,MEM_COMMIT+MEM_RESERVE,PAGE_EXECUTE_READWRITE
     mov     [shd_buffer_ptr],eax
}


macro prompt arg1
{
    push   arg1
    call   prompt_proc
}

macro cls
{
    pushad
    movzx  eax,[xres_var]   ; clear text with render metod
    invoke StretchDIBits,[hdc],0,0,eax,30,0,0,2,2,fakebitmap,Wbmi,0,SRCCOPY
    popad
}



section '.text' code readable executable
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; start and message loop
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
start:

    invoke GetModuleHandle,0
    mov [wc.hInstance],eax
    invoke LoadIcon,0,IDI_APPLICATION
    mov [wc.hIcon],eax
    invoke LoadCursor,0,IDC_ARROW
    mov [wc.hCursor],eax

    mov     eax,1
    cpuid
    ; test bit no 28 of ecx if set - avx on board
    bt	    ecx,28
    jnc     @f
    mov     [isAVX],1  ; .avx_present
  @@:
    bt	    ecx,20
    jnc     @f
    mov     [isSSE42],1
    jmp     .ns4
  @@:
   ; if     Ext <> SSE3
   ; mov     [maxffd],1    ; switch off sse4 options
   ; end     if
    mov     [max_rpbump],1
    mov     [max_draw],19  ; switch off see4 render models
  .ns4:
    mov     edi,ThreadMaskArray
    xor     eax,eax
    mov     ecx,64
    cld
    rep     stosd


 ;   rdtsc
 ;   mov  [rand_seed],ax
	mov	[rand_seed],0xa1f5
	invoke	GetCurrentProcess
	invoke	GetProcessAffinityMask,eax,ProcessAffinityMask,SystemAffinityMask

	mov	ecx,ThreadMaskArray  ; core count snippet - 32bit version tthsqe's
				     ; routine, thanks !
	mov	ebx,[ProcessAffinityMask]
	xor	edx,edx
     .loop:
	mov	eax,ebx
	dec	ebx
	and	ebx,eax
	xor	eax,ebx
	add	edx,1
	mov	[ecx],eax
	add	ecx,4
	test	ebx,ebx
	jnz	.loop
	mov	[CoresCount],  edx

     invoke  GetSystemMetrics,SM_CXSCREEN
     push    eax
     invoke  GetSystemMetrics,SM_CYSCREEN
     mov     edx,eax
     pop     eax

     mov     [initXsize],eax
     mov     [initYsize],edx

     do_size_vars


;    fix font and whole menu


    mov     esi,menu_data
    xor     eax,eax
    mov     ebx,20
    cmp     [yres_var],900
    cmovb   eax,ebx
    cmp     ax,[esi+8]
    jne     .skip_fix
    or	    edi,-1
;    mov     ebx,20
  @@:
    movzx   eax,word[esi+8]
    cdq
    div     ebx
    imul    eax,13
    sub     eax,11
 ;   inc     eax
    mov     [esi+4],ax
    add     eax,12
    mov     [esi+8],ax
    add     esi,rotary_flag.op	 ; size of one menu position
    cmp     [esi],di
    jne     @b

    mov     eax,MXRES * 4 * FONT_FIX
    sub     [bbuffer_ptr],eax
    sub     eax, 40 * MXRES
    sub     [inf_screen_ptr],eax
    sub     [next_var], FONT_FIX

 .skip_fix:

     alloc_screen_mem


     mov     eax,FUR_SIZE*(FUR_SHELLS+1) + 6553
     malloc  eax

    ; fur routines need reedit
     mov     [furs_ptr],eax

    mov      eax,65536*6  ; mem for long pipe size determines with Lp variables
    malloc   eax
    mov      [long_pipe_vert_ptr],eax  ; mem for long pipe vertices

macro do_bmi

{
    mov   [bmi.biSize],sizeof.BITMAPINFOHEADER
    movzx eax,[xres_var]
    mov   [bmi.biWidth],eax  ;XRES
    movzx edx,[yres_var]

;    add   edx,30
    neg   edx
    xor  esi,esi
    mov  [bmi.biHeight],edx ;-YRES-30
    mov  [bmi.biPlanes],1
    mov  [bmi.biBitCount],32
    mov  [bmi.biCompression],BI_RGB
    mov  [bmi.biSizeImage],esi
    mov  [bmi.biXPelsPerMeter],esi
    mov  [bmi.biYPelsPerMeter],esi
    mov  [bmi.biClrUsed],esi
    mov  [bmi.biClrImportant],esi

}
    do_bmi

;fake bitmap bmi
    mov   [Wbmi.biSize],sizeof.BITMAPINFOHEADER
    mov   ebx,2
    mov   [Wbmi.biWidth],ebx  ;XRES
    neg   ebx
    mov   [Wbmi.biHeight],ebx ;-YRES-30
    mov   [Wbmi.biPlanes],1
    mov   [Wbmi.biBitCount],32
    mov   [Wbmi.biCompression],BI_RGB
    mov   [Wbmi.biSizeImage],esi
    mov   [Wbmi.biXPelsPerMeter],esi
    mov   [Wbmi.biYPelsPerMeter],esi
    mov   [Wbmi.biClrUsed],esi
    mov   [Wbmi.biClrImportant],esi


macro menu_bmi

{

    mov    [Mbmi.biSize],sizeof.BITMAPINFOHEADER
    mov    [Mbmi.biWidth],MXRES
    mov    ebx,MYRES + IN_CNST
    movzx  eax,[yres_var]
    cmp    ebx,eax
    cmova  ebx,eax
    neg    ebx
    mov    [Mbmi.biHeight],ebx	;-(MYRES + IN_CNST ) ;-30 ;-BYRES-30
    mov    [Mbmi.biPlanes],1
    mov    [Mbmi.biBitCount],32
    mov    [Mbmi.biCompression],BI_RGB
    xor    eax,eax
    mov    [Mbmi.biSizeImage],eax
    mov    [Mbmi.biXPelsPerMeter],eax
    mov    [Mbmi.biYPelsPerMeter],eax
    mov    [Mbmi.biClrUsed],eax
    mov    [Mbmi.biClrImportant],eax

}

    menu_bmi
    xor eax,eax
    mov [Pbmi.biSize],sizeof.BITMAPINFOHEADER
    mov [Pbmi.biWidth],PROMPTX
    mov [Pbmi.biHeight],-PROMPTY
    mov [Pbmi.biPlanes],1
    mov [Pbmi.biBitCount],32
    mov [Pbmi.biCompression],BI_RGB
    mov [Pbmi.biSizeImage],eax
    mov [Pbmi.biXPelsPerMeter],eax
    mov [Pbmi.biYPelsPerMeter],eax
    mov [Pbmi.biClrUsed],eax
    mov [Pbmi.biClrImportant],eax



if 0   ;  dib bmi


      mov  [bmiheader.biSize],sizeof.BITMAPINFOHEADER
	       movzx  eax,[xres_var]
	       mov    [bmiheader.biWidth],eax
	       movzx  eax,[yres_var]
	       neg    eax
	       mov    [bmiheader.biHeight],eax
	       mov    [bmiheader.biPlanes],1
	       mov    [bmiheader.biBitCount],32
	       mov    [bmiheader.biCompression],BI_RGB
	       movzx  eax,[xres_var]
	       movzx  ebx,[yres_var]
	       imul   eax,ebx
	       lea    eax,[4*eax+16]
	       mov    [bmiheader.biSizeImage],eax
	       mov    [bmiheader.biXPelsPerMeter],0
	       mov    [bmiheader.biYPelsPerMeter],0
	       mov    [bmiheader.biClrUsed],0
	       mov    [bmiheader.biClrImportant],0
	       invoke CreateDIBSection,0,bmiheader,0,screen_ptr,0,0
end if
    xor    ebx,ebx
    mov    [rect.left],ebx
    mov    [rect.top],ebx
    movzx  eax,[whole_xres]
    mov    [rect.right],eax
    movzx  edx,[whole_yres]
    mov    [rect.bottom],edx

    invoke AdjustWindowRectEx,rect,WS_POPUP or WS_SYSMENU or WS_CAPTION,0

    invoke RegisterClass,wc
    xor    ebx,ebx
    mov    ecx,[rect.bottom]
    mov    eax,[rect.right]
    sub    ecx,[rect.top]
    sub    eax,[rect.left]
    ; sizable window
    invoke CreateWindowEx, ebx,_class, _title, WS_VISIBLE + WS_OVERLAPPEDWINDOW,\
				ebx, ebx, eax, ecx, NULL, NULL, [wc.hInstance], NULL

    mov    [hwnd],eax
    invoke GetDC,[hwnd]
    mov    [hdc],eax
    invoke ShowWindow,[hwnd],SW_SHOW

    ; fix bezier rect files
     mov   ebx,tspoon_rect
     mov   ecx,16+26+32 + 6 ; spoon + cup + pot + ffd
     shl   ecx,4	    ; ffd - free form deformation..
   @@:
     dec   word[ebx]
     add   ebx,2
     loop  @b

   ;  call     do_fog_seeds

     push     dword tspoon_nodes
     push     dword 256
     call     normalize_object1

     push     dword tcup_nodes
     push     dword 251
     call     normalize_object1

     push     dword tpot_nodes
     push     dword 306
     call     normalize_object1

     push     dword ffd_nodes
     push     dword 56
     call     normalize_object1


     mov      eax,ffd_nodes
     mov      ecx,56
     mov      edx,0x80000000
     movd     xmm1,edx
     pshufd   xmm1,xmm1,0
   @@:
     movups   xmm0,[eax]
     shufps   xmm0,xmm0,11110110b
     xorps    xmm0,xmm1
     movlps   [eax+4],xmm0
     add      eax,12
     loop     @b



     mov      esi,ffd_nodes
     mov      edi,ffd_nodes_intial
     mov      ecx,56 * 3
     cld
     rep      movsd

     call     long_pipe_init ; init control points of long pipe

			      ; dependent on segments count
     movzx    eax,[yres_var]  ;YRES / 2
     shr      eax,1
     cvtsi2ss xmm0,eax
     movss    [scale],xmm0
     call     normalize_plane_equation
     call     open_file

     call     active_object_do

    call   normalize_all_light_vectors
    mov    edi,lights_aligned
    mov    ecx,3
    call   norm_aligned_lv
    call   do_color_buffer
    call   init_envmap2
    mov    edi,bump_map
    call   calc_bumpmap
    call   generate_texture2
    call   init_point_lights

    mov    ecx,FUR_SIZE
    mov    esi,FUR_SHELLS+1
    mov    edi,[furs_ptr]
    call   fur_shell_generator


    fninit
    fld    [one_deg]
    mov    ecx,256
    mov    edi,sin_cos_tab256
    fldz
  @@:
    fld    st
    fsincos
    fstp   dword[edi]
    fstp   dword[edi+4]
    fadd   [one_deg]
    add    edi,8
    loop   @b
    ffree  st0
    xor    eax,eax
    call   draw_menu

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;   msgLoop
;;;;;;;;;;;;;;;;;;;;;;;;;7;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
msg_loop:
    cmp     [ray_shd_flag],1
    je	    @f
    call    rend_com
    @@:
    cmp     [bezier_flag],0
    jne     @f
    cmp     [edit_flag],0
    jne     @f
    cmp     [speed_flag],1
    je	    @f
    cmp     [lpipe_flag],0
    jne     @f
    invoke  Sleep,8
  @@:


    invoke  PeekMessage,msg,0,0,0,PM_NOREMOVE
    or	    eax,eax
    je	    msg_loop

    invoke  GetMessage,msg,NULL,0,0
    or	    eax,eax
    je	    end_loop

    invoke  TranslateMessage,msg
    invoke  DispatchMessage,msg
    jmp     msg_loop

.error:
    invoke  MessageBox,NULL,_error,NULL,MB_ICONERROR+MB_OK

end_loop:


    invoke  ReleaseDC,[hwnd],[hdc]
    invoke  DestroyWindow,[hwnd]
    invoke  ExitProcess,[msg.wParam]

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; WindowProc
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
proc WindowProc uses ebx esi edi, hwnd,wmsg,wparam,lparam

locals
     .mxx	 rd 40
     .defwindow  rd 1
     .coords	 rd 1
     .end_var	 rd 1
     .calc_combo rd 1
     .triangles_ptr rd 1
     .points_r_ptr    rd 1
endl
     mov	eax,.defwndproc
     mov	edx,.end
     mov	ecx,calc_combo
     mov	[.defwindow],eax
     mov	[.end_var],edx
     mov	[.calc_combo],ecx
     mov	[.triangles_ptr],triangles_ptr
     mov	[.points_r_ptr],points_r_ptr


     cmp       [wmsg],WM_KEYDOWN
     je        .wmkeydown
     cmp       [wmsg],WM_PAINT
     je        .wmpaint
     cmp       [wmsg],WM_SIZE
     je        .wmsize
     cmp       [wmsg],WM_LBUTTONDOWN
     je        .wmlbuttondown
     cmp       [wmsg],WM_MOUSEMOVE
     je        .wmmousemove
     cmp       [wmsg],WM_LBUTTONUP
     je        .wmlbuttonup
     cmp       [wmsg],WM_RBUTTONDOWN
     je        .wmrbuttondown
     cmp       [wmsg],WM_RBUTTONUP
     je        .wmrbuttonup
     cmp       [wmsg],WM_DESTROY
     je        .wmdestroy
     jmp       [.defwindow]


   .wmsize:
     mov       eax,[lparam]
     mov       edx,eax
     shr       edx,16
     cwde
     add       edx,30



if 0
     cmp       [wmsg],WM_SIZING
     jne       .wmp
     mov       esi,[lparam]
     mov       eax,[esi+8]
     mov       edx,[esi+12]
     add       edx,30
     mov       ecx,[esi]
     mov       ebx,[esi+4]

    mov        [rect.left],ecx
    mov        [rect.top],ebx
;    movzx     eax,[whole_xres]
    mov        [rect.right],eax
;    movzx     edx,[whole_yres]
    add        edx,30
;    neg       edx
    mov        [rect.bottom],edx ;YRES+30
    push       esi
    invoke     AdjustWindowRectEx,rect,WS_POPUP or WS_SYSMENU or WS_CAPTION,0
    pop        esi
    mov        eax,[esi+8]
    mov        edx,[esi+12]

end if


     mov       edi,minWinX
     cmp       eax,[edi]
     cmovb     eax,[edi]
     cmp       edx,[edi+4]
     cmovb     edx,[edi+4]
;     cmp	eax,[minWinX]
;     cmovb	eax,[minWinX]
;     cmp	edx,[minWinY]
;     cmovb	edx,[minWinY]
     xor       ecx,ecx
     mov       ebx,1
     cmp       [initXsize],eax
     cmovg     ecx,ebx
     cmp       [initYsize],edx
     cmovg     ecx,ebx
     push      ecx
     do_size_vars  ;macro
     pop       ecx

;     or	ecx,ecx
;     jz	@f
;     mov	[initXsize],eax
;     mov	[initYsize],edx
; realloc "screen" mem only if size greater than in start
;     call	free_screen_mem
;     alloc_screen_mem	;macro
   @@:
     do_bmi    ;macro
     menu_bmi
     clrscr    ;macro
     jmp       [.defwindow]



   .wmpaint:
      ; redraw all
     clrscr
     movzx     eax,[xres_var]
     movzx     ebx,[yres_var]
     invoke    StretchDIBits,[hdc],0,30,eax,ebx,0,0,eax,ebx,[screen_ptr],bmi,0,SRCCOPY
     call      write_info
     call      NextMdraw_all
     xor       eax,eax
     call      draw_menu
     jmp       [.defwindow]


    .wmlbuttondown:
	mov	eax,[lparam]
	sub	eax,30 shl 16
	mov	[.coords],eax
	xorps	xmm0,xmm0
	movd	xmm1,eax
	movd	xmm2,[xres_vard]
	pcmpgtw xmm0,xmm1
	pcmpgtw xmm2,xmm1
	pxor	xmm2,xmm0
	movd	ebx,xmm2
	cmp	ebx,-1
	jne	.nxx_edit
	mov	ebx,eax
	shr	ebx,16 ; y
	cwde
	movzx	edx,word[xres_var]
	imul	ebx,edx
	add	ebx,eax
	shl	ebx,2
	add	ebx,[edit_buf_ptr]
	mov	ecx,dword[ebx]
	cmp	ecx,-1
	jz	.nxx_edit
	mov	eax,ecx
	inc	ecx    ; prevent long pipe cause ?  ;
	cmp	[ffd_flag],0
	jne	@f
	cmp	[bezier_flag],0
	jne	@f
	mov	[edit_particle_no],ecx ;if vert_edit_no = 0, no vertex selected
	xor	edx,edx
	mov	[curr_chunk],eax
	jmp	.f1
      @@:
	dec	ecx
	mov	[ed_bpatch_derive_no],cx
    .f1:
    @@:
     xor      edx,edx
     cmp      [mirr_flag],dl
     je       @f
     call     mirror_chunk
     jmp      .prep_c
   @@:
     cmp      [del_chn_flag],dl
     je       @f
     call     delete_chunk
     jmp      .prep_c
   @@:
     cmp      [invert_n_flag],dl
     je       .nxx_edit
     call     invert_normals_ch

   .prep_c:
     call     remove_dead_tri
     mov      eax,[triangles_ptr]
     mov      ebx,[points_r_ptr]
     mov      ecx,[triangles_count_var]
     mov      edx,[points_count_var]
     call     remove_unused_verticesB
     mov      [points_count_var],ecx

     call     [.calc_combo]
     jmp      [.defwindow]


   .nxx_edit:

     cmp     [ed_bpatch_derive_on],0
     jne     .no_edit_derives
     mov     ebx,[lparam]
     mov     [ed_bpatch_derives_x_start],ebx
     mov     [ed_bpatch_derives_x_end],ebx
     mov     word[ed_bpatch_derive_on],1
  .no_edit_derives:

     cmp     [NextMed_flag],0	; next mesh edition routines
     je      .pipe

     mov     eax,[lparam]
     mov     ecx,[next_var]
     shl     ecx,16

     mov     cx,[xres_var]
     mov     edx,BYRES shl 16 + BXRES
     add     edx,ecx
     movd    xmm0,eax
     movd    xmm1,ecx
     movd    xmm2,edx
     pcmpgtw xmm1,xmm0
     pcmpgtw xmm2,xmm0
     pxor    xmm2,xmm1
     movd    edx,xmm2
     cmp     edx,-1
     jne     .pipe

     movd    xmm1,ecx
     psubw   xmm0,xmm1
     movd    ebx,xmm0

     mov     eax,ebx
     mov     edx,BYRES
     cwde
     shr     ebx,16
     imul    ebx,edx
     add     eax,ebx
     shl     eax,2
     add     eax,[bbuffer_ptr] ;Bbuffer
     mov     ebx,[eax]
     shr     ebx,24
     cmp     bl,cl
     jz      .pipe

  ;   cmp     ebx ->  ; 1 - xy, 2 - z, 3 - scale, 4 - rotate, 5- scale xy,
		      ; 6, 7, 8, 9 - bend

     mov     [NextMotion],bx

   .pipe:
     cmp     [lpipe_flag],1	; try reserve enough mem
     je      .editt
     cmp     [rotary_flag],3
     jne     @f
     mov     eax,[lparam]
     mov     dword[custom_rot_start_x],eax
    @@:
     cmp     [set_tri_area],0
     je      .editt
     cmp     [edit_flag],0
     jne     .editt
     jmp     .no_edit

   .editt:

	movd	xmm1,[lparam]
	mov	ebx,30 shl 16 + 1
	movd	xmm0,ebx
	movaps	xmm4,xmm0
	movd	xmm3,dword[xres_var]
	paddw	xmm3,xmm0
	pcmpgtw xmm0,xmm1
	pcmpgtw xmm3,xmm1
	pxor	xmm3,xmm0
	movd	eax,xmm3
	mov	cx,ax
	shr	eax,16
	and	ax,cx
	or	ax,ax
	jz	.no_edit

	movd	xmm0,[lparam]  ;ebx
	psubw	xmm1,xmm4

	movd	eax,xmm1
	mov	[.coords],eax	; meu coords
    ;	 push	 eax
    ;	 eax - position of cursor in "screen/work" area
	cmp    [set_tri_area],0
	jne    ..edit_mesh ; teslation area work
	movzx  ebx,ax	; check if edition triangulize area
	mov    ecx,eax
	shr    ecx,16
	movzx  edx,word[xres_var]
	imul   edx,ecx
	add    ebx,edx
	mov    ecx,ebx
	shr    ebx,2
	and    ecx,11b
	add    ecx,ecx
	inc    ecx
	mov    edi,edit_tri_area
	add    ebx,[edit_tri_area_ptr]
	bt     [ebx],ecx
	jnc    @f
	mov    dword[edi],1
	jmp    .no_edit
      @@:
	inc    ecx
	bt     [ebx],ecx
	jnc    @f
	mov    dword[edi],2
	jmp    .no_edit
    @@:
      ..edit_mesh:
	mov    edi,edit_particle_no
	mov    eax,[.coords]
	cmp    [edit_flag],2
	jne    @f

	cmp    dword[edi],-1 ;[edit_particle_no],-1
	je     @f
	dec    dword[edi]   ;;[edit_particle_no]
       @@:
	cld
	mov    edi,edit_start_x
	stosd
	stosd
;	 mov	dword[edit_end_x],eax
;	 mov	dword[edit_start_x],eax


   .no_edit:

     mov    eax,[lparam]	; menu work
     movzx  ecx,[xres_var]
     cmp    ax,cx
     jng    .defwndproc
     add    ecx,MXRES
     cmp    ax,cx
     jnl    .defwndproc
     ror    eax,16
     cmp    ax,MYRES
     jnl    .defwndproc
     cmp    [menu_button_number],0
     jne    .defwndproc

     mov    esi,menu_data
     mov    eax,[lparam]
     sub    ax,[whole_xres]
     add    ax,MXRES  ;+MXPRES+10
   @@:
     push   eax
     mov    ecx,[esi+2]
     mov    edx,[esi+6]
     call   xy_in_rect
     pop    eax
     or     ebx,ebx
     jnz    @f
     add    esi,33		 ; size of one position in menu
     cmp    [esi],word -1	 ; end mark
     jnz    @b
     jmp    .st
  @@:
     mov    ax,[esi+2]
     add    ax,2
     shl    eax,16
     mov    ax,[esi+6]
     sub    ax,2
     mov    bx,[esi+4]
     add    bx,2
     shl    ebx,16
     mov    bx,[esi+8]
     sub    bx,2

     mov    edx,MXRES
     movd   xmm2,edx
     movd   xmm7,edx
     mov    edx,menu_screen

     movd   xmm1,edx
     mov    edx,0x0000ff00
     movd   xmm3,edx
     mov    edi,MYRES
     movd   xmm5,edi
     pslldq xmm5,4
     por    xmm7,xmm5
     pxor   xmm6,xmm6
     push   esi

;write_frame:
    ; xmm1  -  buffer ptr
    ; xmm2  -  width
    ; xmm3  -  color
    ; eax  -  x2 shl 16 + x1
    ; ebx  -  y2 shl 16 + y1
    ; xmm6 -  min x, min y
    ; xmm7 -  max x, max y

     call   write_frame
     pop    esi
     mov    word[esi],1      ; mark button chosen
     and    eax,-1	     ; dont clear menu
     call   draw_menu
   .st:
     stretch_menu
     jmp    [.defwindow]



   .wmmousemove:

     stretch_menu
     cmp     [ed_bpatch_derive_on],1
     jne     .Nexf
     push    [lparam]
     pop     dword[ed_bpatch_derives_x_end]
   @@:
 ;   jmp     [.defwindow]
   .Nexf:


     cmp    [NextMed_flag],0   ; next mesh edition routines
     je     .l_pipe
     movzx  ebx,[NextMotion]
     cmp    ebx,0
     je     .l_pipe
     cmp    ebx,9
     jg    .l_pipe
     mov    eax,[lparam]
     sub    ax,[xres_var]
     ror    eax,16
     sub    ax,[next_varw]
     ror    eax,16

     cmp     ebx,1
     jne     @f
     mov     [NextMxadd],eax
     jmp     .callM
   @@:
     cmp     ebx,2
     jne     @f
     shr     eax,16
     mov     [NextMzadd],ax
     jmp     .callM
   @@:
     cmp     ebx,3
     jne     @f
     shr     eax,16
     mov     [NextScale],ax
     jmp     .callM
    @@:
     cmp     ebx,4
     jne     @f
     mov     [NextRotat],eax
     jmp     .callM
    @@:
     cmp     ebx,5
     jne     @f
     mov     [NextScaleXY],eax
     jmp     .callM
    @@:
     cmp     ebx,10	     ; bend
     jg      .l_pipe
     mov     edi,NextBend1
     sub     ebx,6
     shl     ebx,2
     add     edi,ebx
     mov     [edi],eax

   .callM:
     call      write_info
     call      NextMdraw_all
     xor       eax,eax
     call      draw_menu
;     stretch_menu
     jmp       [.defwindow]

   .l_pipe:
     cmp    [lpipe_flag],1
     je     @f
     cmp    [NextMed_flag],0
     jne    @f
     cmp    [rotary_flag],3
     jne    .tri_rect
   @@:

     cmp    [custom_rot_start_x],0
     je     .tri_rect
     movd   xmm0,[lparam]
     movq   xmm1,qword[custom_rot_start_x]
     psubw  xmm0,xmm1
     movd   [custom_obj_angle],xmm0

  .tri_rect:
	cmp	  [edit_tri_area],0
	je	  @f
	mov	  ebx,[lparam]
	movq	  xmm3,[min_work_x]
	movq	  xmm4,[max_work_x]
	movd	  xmm1,ebx
	mov	  ebx,30 shl 16 + 1  ; teslate work
	movd	  xmm0,ebx
	psubw	  xmm1,xmm0
	pmaxsw	  xmm1,xmm3
	pminsw	  xmm1,xmm4
	movdqa	  xmm0,xmm1
	pxor	  xmm1,xmm1
	punpcklwd xmm0,xmm1
	cvtdq2ps  xmm0,xmm0
	cmp	  [edit_tri_area],2
	jnz	  @f
	movlps	  qword[tri_area_x1],xmm0
	jmp	  .end_tr
       @@:
	cmp	  [edit_tri_area],1
	jnz	 .end_tr
	movlps	  qword[tri_area_x2],xmm0
  .end_tr:

   @@:
     cmp     [tex_button_pres],1
     jne     .tx_dn
     movd    xmm0,[lparam]
     movd    xmm1,[tex_mov_begin_x]
     psubw   xmm1,xmm0
     movd    eax,xmm1
     movd    [tex_mov_begin_x],xmm0
     mov     ebx,eax
     sar     ebx,16
     cwde
     call    move_texture
     jmp     [.defwindow]

  .tx_dn:
     or      eax,-1
     cmp     [edit_particle_no],eax    ;0   last_chg
     jz      .next_one

   @@:

     movd    xmm0,[lparam]
     movq    xmm2,[min_work_x]
     movq    xmm3,[max_work_x]
     mov     eax,30 shl 16 + 0
     movd    xmm1,eax
     psubw   xmm0,xmm1
     pmaxsw  xmm0,xmm2
     pminsw  xmm0,xmm3
     movd    dword[edit_end_x],xmm0

  .next_one:

     jmp	[.defwindow]



  .wmlbuttonup:

  ;   stretch_menu
     xor   eax,eax
     call  draw_menu
     cmp   [ed_bpatch_derive_on],1
     jne   .no_bez_submit
   @@:

	cvtdq2ps xmm3,[xxadd]
	mov	 esi,matrix_scaled
	lea	 edi,[.mxx]
	call	 reverse_mx_3x3
	movzx	 eax,[ed_bpatch_derive_no]
	imul	 eax,12
	mov	 edi,ffd_nodes
	mov	 esi,ffd_nodes_rot
     ;	 add	  esi,eax
     ;	 add	  edi,eax
	movzx	 edx,[ffd_flag]
	cmp	 edx,0
	je	 @f
	cmp	 edx,2
	je	 @f
	movaps	 xmm1,[ffd_scale]
	rcpps	 xmm1,xmm1
	movups	 xmm0,[esi+eax]
	subps	 xmm0,xmm3
	subps	 xmm0,[ffd_center]
	shufps	 xmm1,xmm1,0
	mulps	 xmm0,xmm1
	movlps	 [edi+eax],xmm0
	movhlps  xmm0,xmm0
	movss	 [edi+eax+8],xmm0
	jmp	 .ffd1
      @@:
	cmp	 [bezier_flag],0    ; edit bezier patch
	cmovne	 edi,[derv_nodes_ptr]

	add	 edi,eax
	add	 esi,eax
	movups	 xmm0,[esi]
	subps	 xmm0,xmm3
	movlps	 [esi],xmm0
	movhlps  xmm0,xmm0
	movss	 [esi+8],xmm0
	mov	 ecx,1
	lea	 ebx,[.mxx]
	call	 rotary
    .ffd1:
	xor	 eax,eax
	mov	 [ed_bpatch_derive_on],ax
	mov	 [ed_bpatch_derives_x_start],eax
	mov	 [ed_bpatch_derives_x_end],eax

   .no_bez_submit:
     xor    eax,eax
     mov    [NextMotion],ax   ; next mesh x y
     cmp    [lpipe_flag],1
     je     @f
     mov    [custom_rot_start_x],ax
   @@:


     mov    [edit_tri_area],al
     cmp    [edit_particle_no],-1    ;0   last_chg
     je     .segment_rot


     cmp    [lpipe_flag],1
     jne    @f
     mov    esi,long_pipe_derv
     mov    edi,long_pipe_derv_rotated
     call   edit_submit

     mov    eax,[tex_scale]
     call   calc_bumpmap_coords  ; bump , tex
     or     eax,-1
     xor    ebx,ebx
     mov    [edit_particle_no],eax
     mov    dword[edit_end_x],ebx
     mov    dword[edit_start_x],ebx
     jmp    .segment_rot
    @@:
     cmp    [ffd_flag],1
     je     @f
     mov    [do_submit],1      ; submit after rendering
    @@:
   .segment_rot:
     ; GUI buttons work

     cmp    [menu_button_number],0
     jne    .defwndproc
     mov    eax,[lparam]     ; check if sth in menu marked
     sub    ax,[whole_xres]
     cmp    ax,-MXRES ;+MXPRES+10)
     jl     .dr_mn
     ror    eax,16
     cmp    ax,MYRES
     jg     .dr_mn
     ror    eax,16
     mov    esi,menu_data
     mov    eax,[lparam]
     sub    ax,[whole_xres]
     add    ax,MXRES ;+MXPRES+10

  @@:
     push   eax
     mov    ecx,[esi+2]
     mov    edx,[esi+6]
     call   xy_in_rect
     pop    eax
     or     ebx,ebx
     jnz    @f
     add    esi,33		 ; size of one position in menu
     cmp    [esi],word -1
     jnz    @b
     jmp    .dr_mn
  @@:
     cmp    word[esi],1
     jne    .dr_mn
     ; this button was pressed (pointed in esi)
     xor    eax,eax		      ; update flags
     mov    word[esi],ax      ; mark
     mov    al,byte[esi+22]
     inc    al
     cmp    al,[esi+21]
     jne    @f
     xor    al,al
   @@:
     mov    [esi+22],al
     mov    ax,[esi+27]
     xor    ebx,ebx
     cld
     mov    edi,menu_button_number
     cmp    [edi],bx
     jne    @f
     stosw
;     mov    [menu_button_number],ax
     xor    eax,eax
     call   draw_menu
     call   buttons
     jmp    [.end_var]
   @@:
   .dr_mn:
     xor    eax,eax
     call   draw_menu
     jmp    [.defwindow]


  .wmrbuttondown:

    cmp   [lpipe_flag],1
    jne   @f
    mov   eax,[lparam]
    mov   dword[custom_rot_start_x],eax
  @@:
    mov   [tex_button_pres],1
    mov   eax,[lparam]
    mov   [tex_mov_begin_x],eax
    jmp   [.defwindow]



  .wmrbuttonup:
    xor   eax,eax
    mov   [tex_button_pres],al
    cmp   [lpipe_flag],1
    jne   @f
    mov   [custom_rot_start_x],ax
   @@:

  .defwndproc:
    invoke  DefWindowProc,[hwnd],[wmsg],[wparam],[lparam]
    jmp     .finish

  .wmkeydown:
    cmp   [wparam],VK_UP
    jne   @f
    xor   eax,eax
    jmp   [.defwindow]
  @@:
    xor    ebx,ebx
    cmp    [menu_button_number],bx
    jne    @f
    mov    eax,[wparam]
    call   update_flags_after_keypress
    call   buttons
  @@:
    cmp    [wparam],0x4d  ;  m
    jne    @f
    call   zero_flags
    xor    eax,eax
    call   open_next_mesh
    mov    [NextPointsCount],eax
    mov    [NextTrianglesCount],ebx
    mov     eax,2  ; calc edges, use chunks to normals of tri calc
    call    prepare_object
    or	    eax,-1
    mov     [curr_chunk],eax
    call    write_info
    call    draw_menu

   @@:

    cmp     [wparam],0x4e  ;  n
    jne     @f
    call    zero_flags
    or	    eax,-1
    call    open_next_mesh
    mov     eax,2  ; calc edges, use chunks to normals of tri calc
    call    prepare_object
    or	    eax,-1
    mov     [NextPointsCount],0
    mov     [NextTrianglesCount],0
    mov     [curr_chunk],eax
    call    write_info
    xor     eax,eax
    call    draw_menu
    jmp     [.defwindow]
   @@:

    cmp     [wparam],0x58 ;  'x'
    jne     @f
    call    open_texture
    mov     esi,[File_tex_ptr]
    mov     ecx,[FileSize]
    or	    esi,esi
    jz	    @f
    call    read_texture
  @@:

    cmp     [wparam],VK_ESCAPE
    jne     .finish



  .wmdestroy:
    invoke  PostQuitMessage,0
    xor     eax,eax

  .finish:
;    call    buttons
  .end:
    ret
endp


buttons:
  ;===================================================================
  ;==================================================================
  ;===================================================================
  ;==================================================================
  ;===================================================================
  ;===================================================================
  ;======================== menu GUI buttons==========================
  ;===================================================================
  ;===================================================================
    push    ebp
    mov     ebp,esp
    sub     esp,100

    .end_var	equ dword[ebp-4]
    .calc_combo equ dword[ebp-8]


    .points_count_var	   equ dword[ebp-48]
    .triangles_count_var   equ dword[ebp-44]
    .points_r_ptr	   equ dword[ebp-40]
    .triangles_ptr	   equ dword[ebp-36]
    .points_rotated_ptr    equ dword[ebp-32]
    .pnr		   equ dword[ebp-28]
    .edges_ptr		   equ dword[ebp-24]

    .write_info 	     equ dword[ebp-52]
    .remove_unused_verticesB equ dword[ebp-56]
    .remove_dead_tri	     equ dword[ebp-60]
    .prepare_object	     equ dword[ebp-64]

    mov     .end_var,.end
    mov     .calc_combo,calc_combo
    mov     .write_info,write_info
    mov     .remove_unused_verticesB,remove_unused_verticesB
    mov     .remove_dead_tri,remove_dead_tri
    mov     .prepare_object, prepare_object

    cld
    lea      esi,[points_count_var]
    lea      edi,.points_count_var
    mov      ecx,7
    rep      movsd


    movzx   edx,[menu_button_number]
    or	    edx,edx
    jz	    .end
    dec     edx
    add     edx,edx
    lea     esi,[menu_jmp_table_ptr]
    movzx   edx,word[esi+edx]
    or	    edx,edx
    jz	    .end
    add     edx,buttons
    jmp     edx


  .nx55:
    call    sort_chunks
    call    opt_chunks
    mov     eax,[tex_scale]
    call    calc_bumpmap_coords

; do edges list
; in  ->
;	 eax - ptr to edges list in memory (there must be
;	       declared enough size of mem)
;	 ebx - ptr to faces
;	 ecx - faces(triangles) count
;
; out ->
;	 desired list, each edge in list as two dwords - indices
;	   two position in vertices list, pointed by [points_r_ptr]
;	 ecx - edges count

    mov   eax,[edges_ptr]
    mov   ebx,[triangles_ptr]
    mov   ecx,[triangles_count_var]
    call  do_edges_list
    mov   [edges_count],ecx
    jmp   .end_var

  .nx54:
    call    make_series
    call    .calc_combo
    jmp     .end_var


  .nx53:
     call    fix_normals
     mov     eax,1   ;eax  -> not  calc edges
     call    .prepare_object
     jmp     .end_var

  .nx52:
     movss   xmm0,[tex_scale]
     mulss   xmm0,[scale_plus]
     movss   [tex_scale],xmm0
     movd    eax,xmm0
     call    calc_bumpmap_coords
     jmp    .end_var

  .nx51:
     movss   xmm0,[tex_scale]
     mulss   xmm0,[scale_minus]
     movss   [tex_scale],xmm0
     movd    eax,xmm0
     call    calc_bumpmap_coords
     jmp      .end_var

   .nx48:
    ; try to remove cracks
    ; Fix sitution when in the middle of edge ocurr vertex.
    ; This cause arise when invalid we have wrong declaration
    ; of triangles in tris net. In my implementation some numeric
    ; errors could arise :(
    ; To do - not only mid vertex calculation, but every vert along edge.
    ; I tried do above - numerical errors I cannot accept..

  ;  xor     eax,eax ; -> calc edges too
  ;  call    prepare_object
    call    zero_flags	; zero colided edges and chunks count
    call    rem_cracks
    call    .calc_combo
    call    .write_info
    or	    eax,-1
    call    draw_menu
    jmp     .end_var

  .nx47:

     call    .remove_dead_tri
     mov     eax,.triangles_ptr
     mov     ebx,.points_r_ptr
     mov     ecx,[triangles_count_var]
     mov     edx,.points_count_var
     call    .remove_unused_verticesB
     mov     [points_count_var],ecx
     xor     eax,eax ; -> calc edges too
     call    .prepare_object
     call    zero_flags  ; zero colided edges and chunks count
     call    .write_info
     jmp     .end_var


  .nx46:
    cmp     [mark_coll_ed_flag],0
    je	    .end
    cmp     [inner_vert_flag],0
    je	    .end
    mov     ebx,[edges_intersect_ptr]
    mov     esi,[inner_vert_ptr]
;   delete triangles with inside vertices and without intersecting edges
; in  esi - ptr to inside vertices mask list
;     ebx - ptr to intersecting edges mask list
    call    del_tiv_ie_without
    mov     eax,[triangles_ptr]
    mov     ebx,[points_r_ptr]
    mov     ecx,[triangles_count_var]
    mov     edx,[points_count_var]
    call    .remove_unused_verticesB
    mov     [points_count_var],ecx
    xor     eax,eax
    call    .prepare_object
    call    zero_flags	; zero colided edges and chunks count
    call    .write_info
  ;  call    draw_menu
    jmp     .end_var

 .nx45:
   ; inside vertices
    cmp     [inner_vert_flag],0
    je	    .end
    cmp     [inner_vert_ptr],0
    je	    @f
    invoke  VirtualFree ,[inner_vert_ptr] ,0 , MEM_RELEASE
   @@:
    call    mark_inner_vert
    mov     [inner_vert_number],ecx
   ; if vertex is insde adjacent bit is zeroed
    mov     [inner_vert_ptr],ebx
    call    .write_info
    jmp     .end_var

  .nx44:
; teselate triangles with at last one inner vertex
; and with edge that intersect
; in: esi - ptr to inside vertices mask list
;     ebx - ptr to intersecting edges mask list
;     ecx - colided edges number
;     edx - inner vertices number


    xor     ebx,ebx
    cmp     [inner_vert_ptr],ebx
    je	    .end
    cmp     [edges_intersect_ptr],ebx
    je	    .end

  ;  mov     esi,[inner_vert_ptr]
  ;  mov     edx,[edges_intersect_ptr]
  ;  mov     ebx,[edges_coll_count]
  ;  mov     ecx,[inner_vert_number]
    mov     eax,'tiv'
    call    triangulize_all_faces

; out:		 esi - new triangles ptr
;		 edi - new vertices ptr
;		 ecx - new vertices count
;		 ebx - new tris count
 ;   push    ecx
    mov     edx,ecx
    mov     ecx,ebx ;[triangles_count_var]
		    ;  mov     edx,ecx ;[points_count_var]
;    pop     edx
    mov     eax,esi ;[triangles_ptr]
    mov     ebx,edi ;[points_r_ptr]
    push    esi edi ecx
    call    .remove_unused_verticesB
    mov     [points_count_var],ecx

    invoke  VirtualFree ,[points_r_ptr],0 , MEM_RELEASE
    invoke  VirtualFree ,[triangles_ptr],0 , MEM_RELEASE
    pop     [triangles_count_var] [points_r_ptr]  [triangles_ptr]
    xor     eax,eax
    call    alloc_mem_for_tp1

    mov     eax,2
    call    .prepare_object
    mov     ebx,matrix_scaled  ; mx unchanged
    call    rotate_pointsA ;
    call    translate_points
    call    .remove_dead_tri

    mov     eax,[edges_ptr]
    mov     ebx,[triangles_ptr]
    mov     ecx,[triangles_count_var]
    call    do_edges_list
    mov     [edges_count],ecx

    call    rem_cracks
    mov     eax,[triangles_ptr]
    mov     ebx,[points_r_ptr]
    mov     ecx,[triangles_count_var]
    mov     edx,[points_count_var]
    call    .remove_unused_verticesB
    mov     [points_count_var],ecx

    call    .remove_dead_tri
    call    zero_flags	; zero colided edges and chunks count
    call    .write_info
    call    draw_menu
    jmp     .end_var

  .nx43:
    cmp     [ray_shd_flag],1		; ray casted shadows
    jne     .end
    movzx   eax,[draw_flag]
    push    eax
    prompt  prompt_ray_sh
    mov     [draw_flag],100
    stdcall g_draw
    pop     eax
    mov     [draw_flag],al
    movzx   eax,[xres_var]
    movzx   ebx,[yres_var]
    invoke  StretchDIBits,[hdc],0,30,eax,ebx,0,0,eax,ebx,[screen_ptr] ,bmi,0,SRCCOPY
    cls
    prompt  prompt_ray_shd
    jmp     .end_var


  .nx41:

  ;  crop front faces

    push    .points_rotated_ptr
    push    .points_count_var
    call    normalize_object1

    or	    ebx,-1
    mov     edi,[points_rotated_ptr]
    call    rm_inner_faces_th
    mov     eax,[triangles_ptr]
    mov     ebx,[points_r_ptr]
    mov     ecx,[triangles_count_var]
    mov     edx,[points_count_var]
    call    .remove_unused_verticesB
    mov     [points_count_var],ecx
    call    .calc_combo
    jmp     .end_var


   .nx39:
   ;  clip faces
    call    clip_triangles  ;  clipping faces in teslate area

    mov     eax,[triangles_ptr]
    mov     ebx,[points_r_ptr]
    mov     ecx,[triangles_count_var]
    mov     edx,[points_count_var]
    call    .remove_unused_verticesB
    mov     [points_count_var],ecx

    call    .calc_combo
    mov     eax,[triangles_ptr]
    mov     ebx,[points_r_ptr]
    mov     ecx,[triangles_count_var]
    mov     edx,[points_count_var]
    call    .remove_unused_verticesB
    mov     [points_count_var],ecx
    jmp    .end_var

  .nx38:

   ; smooth inside edges
    mov    eax,1   ;   smooth inside edges
    call   mark_colided_edges  ;  try to smooth inside edges
    xor    eax,eax   ;	    - edges mark
    call   .prepare_object
    jmp    .end_var

  .nx35:
 ; remove  inner faces
    mov    edi,.points_r_ptr
    call   rm_inner_faces
    mov    eax,[triangles_ptr]
    mov    ebx,[points_r_ptr]
    mov    ecx,[triangles_count_var]
    mov    edx,[points_count_var]
    call   .remove_unused_verticesB
    mov    [points_count_var],ecx
    call   .calc_combo
    jmp    .end_var

   @@:
   .nx37:
   ; detect chunks
    cmp     [chunks_ptr],0
    je	    @f
    cmp     [chunks_count],0
    je	    @f
    invoke  VirtualFree,[chunks_ptr] , 0,  MEM_RELEASE
   @@:
    xor     eax,eax
    call    detect_chunks
    mov     [chunks_ptr],ebx
    mov     [chunks_count],ecx
    call    .write_info
    jmp     .end_var



 .nx36:
   ; detect intersecting edges

    cmp     [edges_intersect_ptr],0
    je	    @f
    mfree   [edges_intersect_ptr]
   @@:
    cmp     [mark_coll_ed_flag],0
    je	    .end  ;.nx32
    mov     ebx,[edges_count]
    shr     ebx,3
    add     ebx,32
    malloc  ebx
    mov     [edges_intersect_ptr],eax
    mov     edi,eax
    mov     ecx,ebx
    shr     ecx,2
    cld
    xor     eax,eax
    rep     stosd
    mov     [edges_coll_count],eax

    xor     eax,eax   ; - no smooth inside edges
    call    mark_colided_edges	; or try to smooth inside edges
    call    .write_info
    jmp    .end_var





   .nx32:
  ; remove  redundant  vertices
      movzx	ebx,[tolerancy_flag]
      inc	ebx
      cvtsi2ss	xmm0,ebx
      mov	eax,2
      cvtsi2ss	xmm1,eax
      mulps	xmm0,xmm1
      movd	eax,xmm0
      push	.points_count_var
      mov	ebx,11b
      mov	ecx,111b
      cmp	[Z_care_flag],0
      cmovne	ebx,ecx 	  ; button to non chck if in teslate are to do!!
      xor	ecx,ecx 	  ;  check if in teslate area
      call	remove_redundant_vert
      pop	ecx
      cmp	ecx,[points_count_var]
      je	.end
      mov	eax,2
      call	.prepare_object
      jmp	.end_var

  .nx26:
  ; remove unused vertices
    mov   eax,.triangles_ptr
    mov   ebx,.points_r_ptr
    mov   ecx,.triangles_count_var
    mov   edx,.points_count_var
    call  .remove_unused_verticesB
    mov   [points_count_var],ecx
    call  .remove_dead_tri
    xor   eax,eax
    call  .prepare_object
    jmp   .end_var



   .nx31:

   ; edit  next mesh

;    cmp     [NextMed_flag],3
;    jz      @f
;    xor     eax,eax
;    mov     [NextPointsCount],eax
;    mov     [NextTrianglesCount],eax
    call    .write_info
    call    NextMdraw_all
    stretch_menu
    jmp    .end_var

  .nx3:
  ; zoom in
    movss xmm0,[scale]
    mulss xmm0,[scale_plus]
    movss [scale],xmm0
    jmp   .end_var

  .nx4:
  ; zoom out
    movss xmm0,[scale]
    mulss xmm0,[scale_minus]
    movss [scale],xmm0
    jmp   .end_var

  .nx5:
  ; tex map
  ; bump, texture - spherical or planar mapping
    mov   eax,[tex_scale]
    call  calc_bumpmap_coords
    jmp   .end_var

  .nx6:
  ; random directional lights
    call  make_random_lights
    call  normalize_all_light_vectors
    call  init_point_lights
    call  do_color_buffer
    call  init_envmap2
    jmp   .end_var

  .nx30:
 ; long pipe derivesv
 ; init work, depend segs cnt.
    movzx eax,[lp_quality]
    add   eax,2
    shl   eax,2
    mov   [long_pipe_one_seg_c],eax
    jmp   .end_var

  .nx28:
 ; long pipe derivesv
 ; init work, depend segs cnt.
    call  long_pipe_init
    jmp   .end_var
;+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  .nx27:
    mov   al,[lpipe_flag]
    cmp   al,0
    je	  .act
    cmp   al,1
    je	  @f
    cmp   al,2
    je	  .lp_flg2

    xor   ebx,ebx
    mov   [NextPointsCount],ebx
    mov   [NextTrianglesCount],ebx
  @@:
    pushad
    push  dword long_pipe_derv ;[derv_nodes_ptr]
    mov   ecx,[long_pipe_segs_c]
    lea   ecx,[ecx*3+1]
    push  ecx
    call  normalize_object1
    popad
    jmp   .end_var


  .lp_flg2:
       call   zero_flags

   ;	xor	eax,eax
       mov     eax,'frea'
       call    free_mem_for_tp1
       mov     [triangles_count_var],20000
       mov     [points_count_var],20000
       mov     eax,'alla'; 2  ;     eax,eax
       call    alloc_mem_for_tp1

       ; calculate long pipe
       mov    esi,[long_pipe_vert_ptr]	 ;[points_rotated_ptr]
       mov    ecx,[long_pipe_segs_c]
       cmp    [lpcurve_tp],0   ; bspline or bezier long pipe ?
       je     @f
       lea    ecx,[ecx*3-2]
       jmp    .bspl
      @@:
       dec     ecx
    .bspl:

       ; calc long pipe
       imul	ecx,[long_pipe_one_seg_c]
       mov	eax,9	   ; rotary count per one turn
       movzx	edx,[tolerancy_flag]
       imul	edx,6
       cvtsi2ss xmm0,edx   ; radius according approximation tolerancy flag
  ;	mov	edx,14.0  ;5.15   ; radius
  ;	movd	xmm0,edx

       mov    ebx,[triangles_ptr]
       mov    edi,[points_r_ptr]
; in:  esi  - ptr to curve vertices - middle points of pipe
;      ecx  - curve vertices count
;      eax  - rotary steps count as integer
;      xmm0 - radius, lowest dword float
;      ebx  - ptr to triangles list
;      edi  - ptr to pipe vertices
; out: ecx  - triangles count
;      ebx  - vertices count

       call   do_long_pipe
       mov    [triangles_count_var],ecx
       mov    [points_count_var],ebx
       mov    eax,2 ;xor	eax,eax  ; with edges
       call   .prepare_object
       call   .write_info
       call   draw_menu
       jmp    .end
    .act:
      call    active_object_do
      xor     eax,eax
      call    draw_menu
      jmp     .end_var

;++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   .nx7:
; displaying models
    cls
    call   do_color_buffer
    jmp   .end_var


  .nx8:
  ; bumps random/texture dependent
    call  calc_bumpmap
    jmp   .end_var


  .nx9:
  ; bumps random/texture dependent
    movzx eax,[tex_flag]
    or	  eax,eax
    jne   @f
    call  generate_texture2    ; draw stripes
    jmp   .tex_end
  @@:
    cmp   eax,1
    jne   @f
    call  generate_texture3	; draw xor
    jmp   .tex_end
  @@:
    call  generate_texture1	; white area
  .tex_end:
    call  calc_bumpmap
    call  init_point_lights
    jmp   .end_var

  .nx10:

    ; do counter work
    push  esi edi ebx


    mov   ecx,[object_cnt]
    mov   eax,Bfile_name
    mov   [eax+6],cl
    mov   eax,asc_file_name
    mov   [eax+6],cl
    mov   eax,Chfile_name
    mov   [eax+6],cl

    inc   ecx
    mov   eax,'A'
    cmp   ecx,'Z'
    cmove ecx,eax

    mov   [object_cnt],ecx

    pop   ebx edi esi


  ; save 3d object
    call  write_asc
    push  [triangles_count_var]    ; pushes prevent crash in
    push  [points_count_var]	   ; above 65535 vertices or tris
    call  save3ds		   ; situation
    pop   [points_count_var]
    pop   [triangles_count_var]
    call  save_chunks_file
    jmp   .end_var

  @@:
  .nx13:
  ; do displacement
    pushad
    call  apply_displacement_from_tex
    popad
    mov   eax,1  ;no edges
    call  .prepare_object
    jmp   .end_var

  @@:
  .nx14:
  ; do teslation
    xor    eax,eax
    call   triangulize_all_faces ;triangulize_faces
; out:		 esi - new triangles ptr
;		 edi - new vertices ptr
;		 ecx - new vertices count
;		 ebx - new tris count


    cmp    [tes_mod_flag],0	 ; check if teslate area ?
    je	   @f			 ; if zero - not 0 t, 1 - w
    cmp    [td_wp_flag],1	 ; positive Z only  ?
    je	   @f

    push    esi edi ebx
    mov     [points_count_var],ecx
    invoke  VirtualFree ,[points_r_ptr],0 , MEM_RELEASE
    invoke  VirtualFree ,[triangles_ptr],0 , MEM_RELEASE
    pop     [triangles_count_var] [points_r_ptr]  [triangles_ptr]

    call   .calc_combo
    jmp    .end_var

  @@:
    mov     edx,ecx
    mov     ecx,ebx
    mov     eax,esi
    mov     ebx,edi
    push    esi edi ecx
    call    .remove_unused_verticesB
    mov     [points_count_var],ecx

    invoke  VirtualFree ,[points_r_ptr],0 , MEM_RELEASE
    invoke  VirtualFree ,[triangles_ptr],0 , MEM_RELEASE
    pop     [triangles_count_var] [points_r_ptr]  [triangles_ptr]

    xor   eax,eax
    call  alloc_mem_for_tp1
    mov   eax,2
    call  .prepare_object
    call  .remove_dead_tri
    call  rem_cracks

    jmp   .end_var


  .nx15:
  ; current object
    call   active_object_do
    jmp    .end_var

  .nx16:
 ; random point light position
 ; and plane equation
    call   random_point_light_position
    call   normalize_plane_equation
    jmp    .end_var

  .nx19:
  ; transparent displacement color
    cmp    [disp_col_flag],0
    jne    @f
    mov    [displ_transparent_col],0x00ffffff
  @@:
    cmp    [disp_col_flag],1
    jne    @f
    mov    [displ_transparent_col],0
  @@:
    jmp    .end_var

  .nx56:   ; submit changes
    call      .write_info
    call      NextMdraw_all
    xor       eax,eax
    call      draw_menu
    mov       [do_submit],1
    jmp       .end_var

  .nx21:

 ; ********check if draw bezier surfaces******************************
 ;********************************************************************
     mov     eax,1
     call    zero_flags
     xor     ebx,ebx
     mov     [NextPointsCount],ebx
     mov     [NextTrianglesCount],ebx
     mov     bl,[bezier_flag]
     or      bl,bl
     jz      .act1
     lea     esi,[bez_patch_count]
     lea     edi,[bez_nodes_count]
     lea     eax,[derv_rect_ptr]
     lea     edx,[derv_nodes_ptr]
   .do_bez:
     cmp     bl,1
     jne     @f
     mov     dword[eax],tpot_rect
     mov     dword[edx],tpot_nodes
     mov     word[esi],32
     mov     word[edi],306
     jmp    .forw_bez
   @@:
     cmp     bl,2
     jne     @f
     mov     dword[eax],tcup_rect
     mov     dword[edx],tcup_nodes
     mov     word[esi],26
     mov     word[edi],251
     jmp     .forw_bez
   @@:
     cmp     bl,3
     jne     @f  ;ive_object_do
     mov     dword[eax],tspoon_rect
     mov     dword[edx],tspoon_nodes
     mov     dword[esi],16
     mov     dword[edi],256
     jmp     .forw_bez
   @@:
     cmp     bl,4
     jne     .act1  ;ive_object_do
     mov     dword[eax],ffd_rect
     mov     dword[edx],ffd_nodes
     mov     word[esi],6
     mov     word[edi],56

   .forw_bez:
     jmp    .end_var


;---------------------------------------------------------------------
;------------------------end bezier surfaces routines-----------------
;---------------------------------------------------------------------

   .act1:
     or    eax,-1
     mov   [edit_particle_no],eax
     mov   [ed_bpatch_derive_no],ax
     call  active_object_do
     jmp   .end_var

  .nx57:

     mov      edi,ffd_nodes
     mov      esi,ffd_nodes_intial
     mov      ecx,56*3
     cld
     rep      movsd

;     mov      esi,ffd_nodes
;     mov      edi,ffd_nodes_rot
;     mov      ecx,16*6*3
;     cld
;     rep      movsd
     lea      esi,[NextMsub]
     lea      edi,[NextMxadd]
     mov      ecx,17
     rep      movsw

  ;   movups  xmm0,[NextBendZero]
  ;   movups  xmm1,[NextMsub]
  ;   movlps  xmm2,[NextRotatZero]

  ;   movups  [NextBend1],xmm0
  ;   movups  [NextMxadd],xmm1
  ;   movlps  [NextRotat],xmm2

     call      .write_info
     call      NextMdraw_all
     xor       eax,eax
     call      draw_menu
     jmp       .end_var

  .nx62:
      call    do_deformation_normals_depend
      call    .calc_combo
      call    .write_info
      or      eax,-1
      call    draw_menu
      jmp     .end_var

  .nx61:
     movzx    eax,[rph_bump_flag]
     cvtsi2ss xmm1,eax
     movlps   xmm0,[norm_dev2]
     shufps   xmm1,xmm1,0
     mulps    xmm0,[f3]
     mulps    xmm0,xmm1
     movlps   [norm_dev],xmm0
     jmp      .end_var

  .nx20:
  ; displacement factor
    movzx ecx,[disp_fac_flag]
    inc   ecx
    mov   eax,3400
    shl   eax,cl
    mov   [displac_div],eax
;   jmp   .end
  .end:
    mov   [menu_button_number],0

   mov	  esp,ebp
   pop	  ebp

ret

;******************************************************
;*****************			***************
;*****************  Various procedures	***************
;*****************			***************
;******************************************************
;=======================================================
; some long pipe work
pipe_A:
       ;   push  ebp
       ;   mov	 ebp,esp
       ;   sub	 esp,10

	  call	     rotate_long_pipe_derv   ; rotate derives, scale
	  or	     edx,-1				      ; and translate on vector
	  mov	     eax,[edit_particle_no]
	  cmp	     eax,edx
	  je	     @f

	  movd	     xmm0,dword[edit_end_x]
	  movd	     xmm1,dword[edit_start_x]
	  pxor	     xmm2,xmm2
	  punpcklwd  xmm0,xmm2
	  punpcklwd  xmm1,xmm2
	  psubd      xmm0,xmm1
	  cvtdq2ps   xmm0,xmm0

	  dec	     eax
	  shl	     eax,2
	  lea	     eax,[eax*3]
	  add	     eax,long_pipe_derv_rotated
	  movlps     xmm1,[eax]
	  addps      xmm1,xmm0
	  movlps     [eax],xmm1

	@@:
	  mov	 edx,[long_pipe_vert_ptr]
	  mov	 eax,[long_pipe_one_seg_c]
	  mov	 ebx,[long_pipe_segs_c]   ;segments count
	  cmp	 [lpcurve_tp],0
	  je	 @f
	  lea	 ebx,[ebx*3]
	@@:
	  mov	 ecx,long_pipe_derv_rotated
	  mov	 eax,[long_pipe_one_seg_c]

	  push	 eax
	  push	 ebx
	  movzx  esi,[lpcurve_tp]
	  call	 calc_long_curve ;Bsplines or bezier curves
	  mov	 [lp_lenght],ecx
	  pop	 ebx
	  pop	 eax

	  imul	 eax,ebx

	  mov	 [points_count_var],eax
	  mov	 [triangles_count_var],3

     ;	   xor	  eax,eax	   ; clear draw buffer
     ;	   mov	  edi,[screen_ptr]
     ;	   movzx  ebx,[xres_var]
     ;	   movzx  ecx,[yres_var]
     ;	   imul   ecx,ebx
     ;	   cld
     ;	   rep	  stosd


	     ;	draw_vertices
	  mov	    esi,[long_pipe_vert_ptr]  ;[points_rotated_ptr]
	  mov	    ecx,[points_count_var]
	  movzx     edx,[xres_var]

	  pxor	    xmm3,xmm3
	  movd	    xmm4,[xres_vard]
	  punpcklwd xmm4,xmm3
	  cmp	    [lpcurve_tp],0
	  je	    @f
	  mov	    eax,[long_pipe_one_seg_c]
	  sub	    ecx,eax
	  sub	    ecx,eax
	@@:

	.again_vertex:
	  movups    xmm0,[esi]
	  cvtps2dq  xmm0,xmm0
	  pxor	    xmm3,xmm3
	  movd	    xmm4,[xres_vard]
	  punpcklwd xmm4,xmm3
	  pcmpgtd   xmm4,xmm0
	  pcmpgtd   xmm3,xmm0
	  pxor	    xmm3,xmm4
	  pmovmskb  eax,xmm3
	  and	    eax,0x000000ff
	  cmp	    eax,0xff
	  jne	    @f


	  movd	   eax,xmm0
	  psrldq   xmm0,4
	  movd	   ebx,xmm0
	  imul	   ebx,edx
	  add	   ebx,eax
	  shl	   ebx,2
	  add	   ebx,[screen_ptr]
	  mov	   [ebx],dword 0x00ffffff
	@@:			    ;  skip vertex
	  add	   esi,12
	  loop	   .again_vertex


	  ; in esi - derives vertices, each derive 4 vertices,
	  ;	    each derive vertex 3xdword float
	  ;    ecx - derives count
	  ;    edi - screen buffer ptr
	  ;    eax - screen width
	  ; derives list must be scaled and translated into screen coords
	  mov	 esi,long_pipe_derv_rotated
	  mov	 ecx,[long_pipe_segs_c]
	  lea	 ecx,[ecx*3]
	  mov	 edi,[screen_ptr]
	  movzx  eax,[xres_var]
	  call	 draw_long_pipe_derv_lines

	  movzx    eax,[xres_var]
	  movzx    ebx,[yres_var]
	  mov	   ecx,[long_pipe_segs_c]
	  lea	   ecx,[ecx*3]
	  mov	   edx,[screen_ptr]
	  mov	   esi,long_pipe_derv_rotated
	  mov	   edi,[edit_buf_ptr]
	  call	   write_edit_bars_vert  ; write handlers to derives

    ;	   mov	    esp,ebp
    ;	   pop	    ebp

ret
;===============================================================
prompt_proc:
; some text info on top bar
    push   ebp
    mov    ebp,esp

    pushad
    push   dword[ebp+8]

    xor    eax,eax	    ; clear draw buffer
    mov    edi,prompt_screen
    mov    ecx,PROMPTX * PROMPTY
    cld
    rep    stosd
;   eax - x * 65536 + y -> where to display
;   ebx - pointer to character table
;   ecx - size x * 65536 + size y -> size of scr buffer
;   esi - pointer to ASCIIZ string to display
;   edi - pointer to scr buffer
    mov    eax,5 shl 16 + 5
    mov    ebx,font_table + 62 ;  /fonttab2 + 62
    mov    ecx,PROMPTX shl 16 + PROMPTY
    pop    esi
    mov    edi,prompt_screen
    mov    edx,16   ; / 9
    call   write_text1
    invoke StretchDIBits,[hdc],1,1,PROMPTX,PROMPTY,0,0,PROMPTX,PROMPTY,prompt_screen,Pbmi,0,SRCCOPY
    popad
    mov    esp,ebp
    pop    ebp
ret 4
;==============================================================
; rendering procedure commands
rend_com:

   push    ebx edi esi
   push    ebp
   mov	   ebp,esp
   sub	   esp,20

   .scr    equ dword[ebp-4]
   .xres   equ dword[ebp-8]
   .yres   equ dword[ebp-12]
   .nextM  equ dword[ebp-16]



   mov	   esi,[screen_ptr]
   movzx   ebx,[xres_var]
   movzx   ecx,[yres_var]
   movzx   edx,[NextMed_flag]
   mov	   .scr, esi
   mov	   .xres, ebx
   mov	   .yres, ecx
   mov	   .nextM, edx
   or	   eax,-1	  ;
   mov	   edi,[edit_buf_ptr]
   mov	   ecx,.yres
   imul    ecx,.xres
   cld
   rep	   stosd

	  cmp	  [blur_flag],1
	  jne	  @f
	  xor	  eax,eax	   ; clear draw buffer
	  mov	  edi,.scr
	  mov	  ecx,.yres
	  imul	  ecx,.xres
	  cld
	  rep	  stosd

	@@:

    cmp     [lpipe_flag],1 ;   long pipe cause
    jne     .no_pipe_edit
    call    pipe_A
    jmp     .no_draw

 .no_pipe_edit:

    stdcall g_draw
    call    write_info
 .no_draw:
    cmp     [edit_flag],0  ; edition - vert, face and edge
    je	    .skip_ed
    call    write_edit_bars


  .skip_ed:
    movzx  eax,[ccounter_inc]
    ; if custom rotate
    mov    esi,main_rotary
    cmp    [lpipe_flag],1
    je	   .custom
    mov    cl,[rotary_flag]
    cmp    cl,3
    je	   .custom
    xor    ebx,ebx
    cmp    [edit_flag],0
    cmovne eax,ebx
    shl    cl,3
    shl    eax,cl
    pxor   xmm6,xmm6
    movd   xmm7,eax
    punpcklbw xmm7,xmm6
    punpcklwd xmm7,xmm6
    jmp       .end_rot_inc


  .custom:	 ; custom rotate
    push  [custom_obj_angle]
    pop   eax
    and   eax,0x00ff00ff
    mov   ebx,eax
    shr   ebx,16
    mov   dword[esi],ebx
    mov   dword[esi+4],eax
    pxor  xmm7,xmm7

  @@:
  .end_rot_inc:

    pcmpeqd xmm0,xmm0
    movdqu  xmm1,[esi]
    paddd   xmm1,xmm7
    psrld   xmm0,24
    pand    xmm1,xmm0
    movhlps xmm2,xmm1
    movlps  [esi],xmm1
    movss   [esi+8],xmm2
; draw image
    mov     eax,.xres
    mov     ebx,.yres
    invoke  StretchDIBits,[hdc],0,30,eax,ebx,0,0,eax,ebx,[screen_ptr] ,bmi,0,SRCCOPY
  ;  invoke  BitBlt,[hdc],0,30,eax,ebx,[screen_ptr],bmi,0,SRCCOPY

    stretch_menu

    movzx   eax,[draw_flag]
    shl     eax,1
    add     eax,shd_model_addr
    movzx   eax,word[eax]
    add     eax,sh_model
    prompt  eax

    mov     esp,ebp
    pop     ebp
    pop     esi edi ebx

ret



;=============================================================================

active_object_do:
    pushad
    xor    eax,eax
    call   zero_flags
    mov    [NextPointsCount],eax ; disable next mesh
    mov    [NextTrianglesCount],eax
    call   zero_flags  ; zero colided edges and chunks count
    call   write_info
    xor    eax,eax
    call   draw_menu

    cmp    [hrt_flag],1
    jne    @f
    call   read_hrt
    xor    eax,eax
    call   prepare_object
    jmp    .end
  @@:
    cmp    [hrt_flag],2
    jne    @f
    call   load_tetrahedron
    xor    eax,eax
    call   prepare_object
    jmp    .end
  @@:
    mov    eax,'frea'	  ; free all
    call   free_mem_for_tp1
    cmp    [file_ext_flag],0
    je	   @f
    call   read_asc_tp_variables	 ; asc
    mov    eax,'alla'	  ; free all, alloc all
    call   alloc_mem_for_tp1
    call   read_asc
    jmp    .prep
  @@:
    mov    ebx,[FileSize]
    mov    eax,[File_ptr]
    xor    edx,edx  ; no chunks read
    call   read_tp_variables1	     ; 3ds
    mov    [points_count_var],ecx
    mov    [triangles_count_var],ebx
    mov    eax,'alla'	 ; free all, alloc all
    call   alloc_mem_for_tp1

; read from 3ds file
;in:  eax - ptr to triangles
;     ebx - ptr to vertices
;     ecx - ptr to file begin
;     edx - ptr to file end
;out: ecx - points (vertices) count
;     edx - triangles count

    mov    eax,[triangles_ptr]
    mov    ebx,[points_r_ptr]
    mov    ecx,[File_ptr]
    call   read_from_file     ; 3ds
    mov    [EndFile],ebx
  .prep:
    xor    eax,eax
    call   calc_combo

   .end:
   popad
ret
;===============================================================
zero_flags:   ; zero some flags
    xor     ebx,ebx
  ;  mov     [NextPointsCount],ebx
  ;  mov     [NextTrianglesCount],ebx
    invoke  VirtualFree,[inner_vert_ptr] , 0,  MEM_RELEASE
    mov     [inner_vert_ptr],ebx
    mov     [inner_vert_flag],bl
    invoke  VirtualFree,[edges_intersect_ptr],0,MEM_RELEASE
    mov     [edges_intersect_ptr],ebx
    mov     [mark_coll_ed_flag],bl
ret


;======================================================================

long_pipe_init: ; calc long pipes control points
		; segments count depend
    pushad
    movzx eax,[lpsegs_flag]
    mov   ecx,eax
    mov   ebx,2
    shl   ebx,cl
    xor   edx,edx
    mov   [long_pipe_segs_c],ebx
    mov   esi,long_pipe_derv_init
    mov   edi,long_pipe_derv+12
    mov   ecx,[long_pipe_segs_c]
    lea   ecx,[ecx*3]
    add   ecx,4
    call  init_long_pipe_derives
    popad

ret
;=================================================================
prepare_object:
;    in:   eax=1 - no edges; no chunks in normal clc
;	  if
;	   eax=3 calc edges; use chunks info in normal vectors calculations
;	     esi - .tri_ch
;	     edi - .t_ptr
;	  if eax = 2 -> calc edges, calc chunks

    push    ebp
    mov     ebp,esp
    sub     esp,20

    .edg    equ dword[ebp-4]
    .tri_ch equ dword[ebp-8]
    .t_ptr  equ dword[ebp-12]

    mov     .edg,eax
    cmp     eax,3     ; chunks method - obligatory
    je	    @f
    invoke  VirtualFree, [chunks_ptr] , 0,  MEM_RELEASE
    mov     eax,1   ; export  tri_ch and t_ptr
    call    detect_chunks
    mov     [chunks_ptr],ebx
    mov     [chunks_count],ecx
   ;  esi - .tri_ch
   ;  edi - .t_ptr
    mov     eax,3  ; calc edges, use chunks to calc per vertex normal vectors
  @@:

    mov    .tri_ch,esi
    mov    .t_ptr,edi

    pushad
    push   ebp
    push   dword [points_r_ptr]
    mov    ecx,[points_count_var]
    push   ecx
    call   normalize_object1

    call   init_triangles_normals
    pop    ebp
  ;  cmp    [lpipe_flag],2     ; pipes - old/simplfied normal vectors
			       ; calculating metod
  ;  je     .old_m
    cmp    [points_count_var],100  ; below 1000 vertices -
				   ; - use simplified metod, too
    ja	   .new_m
  .old_m:
    mov    ebx,100
    call   init_point_normals
    jmp    .skp
  .new_m:
    mov    edx,3  ;.edg    ; if edx = 3  use chunks lists to computation
    mov    esi,.tri_ch
    mov    edi,.t_ptr
    call   init_pt_norm_th
  .skp:
    popad
    mov    eax,[tex_scale]
    call   calc_bumpmap_coords
    call   sort_chunks
 ;   call  opt_chunks

    cmp   .edg,1
    je	  @f
    mov   eax,[edges_ptr]
    mov   ebx,[triangles_ptr]
    mov   ecx,[triangles_count_var]
    call  do_edges_list
    mov   [edges_count],ecx
  @@:
    mov   esp,ebp
    pop   ebp
ret
;=====================================================================
update_flags_after_keypress:
  ;   in ax - wparam

     pushad
     mov   esi,menu_data
   @@:
     cmp   word[esi+29],ax
     je    .updat
     cmp   word[esi+31],ax
     je    .updat
     add   esi,33
     cmp   word[esi],-1
     jnz   @b
     jmp   .end
  .updat:
     mov   al,byte[esi+22]
     inc   al
     cmp   al,[esi+21]
     jne   @f
     xor   al,al
   @@:
     mov   [esi+22],al
     mov   ax,[esi+27]
     mov   [menu_button_number],ax

     xor    eax,eax
     call   draw_menu
   .end:
     popad
ret

;========================================================

rotate_normals:
   pushad
   push   ebp
   mov	  ebp,esp
   sub	  esp,124

   .nnx_mx  equ [ebp-36]
   .nnx_rot equ [ebp-44]
   .nsinx   equ [ebp-48]
   .ncosx   equ [ebp-52]
   .nsiny   equ [ebp-56]
   .ncosy   equ [ebp-60]
   .nsinz   equ [ebp-64]
   .ncosz   equ [ebp-68]

    .tn 		   equ dword[ebp-124]
    .tnr		   equ dword[ebp-120]
    .pn 		   equ dword[ebp-116]
    .pc 		   equ dword[ebp-112]
    .tc 		   equ dword[ebp-108]
    .points_r_ptr	   equ dword[ebp-104]
    .triangles_ptr	   equ dword[ebp-100]
    .points_rotated_ptr    equ dword[ebp-96]
    .pnr		   equ dword[ebp-92]
    .edges_ptr		   equ dword[ebp-88]
    .nxpc    equ dword[ebp-84]
    .nxtc    equ dword[ebp-80]

    cld
    lea      esi, [triangles_normals_ptr]  ;[points_normals_ptr]  ;[points_count_var]
    lea      edi,.tn
    mov      ecx,10
    rep      movsd


   mov	   ebx,[NextPointsCount]
   mov	   eax,[NextTrianglesCount]
   mov	   .nxpc,ebx
   mov	   .nxtc,eax
   mov	   esi,.pn	     ; rotate normals
   mov	   edi,.pnr
   mov	   ecx,.pc  ;[points_count_var]
   lea	   ebx,[matrix]

   call    rotary

   mov	   esi,.tn   ;[triangles_normals_ptr]		; rotate normals
   mov	   edi,.tnr  ;[triangles_normals_rotated_ptr]
   mov	   ecx,.tc   ;[triangles_count_var]
   lea	   ebx,[matrix]
   call    rotary

    xor      eax,eax
    cmp      .nxpc,eax	 ;.nxpc    ; Next mesh custom scale
    je	     .nno_next

    pxor      xmm3,xmm3
    mov       eax,[NextRotat]
    sub       eax,dword  [NextRotatZero]
    movd      xmm7,eax
    punpcklwd xmm7,xmm3
    movq      .nnx_rot,xmm7

    cld
    lea    esi,.nnx_rot
    lodsd
    mov    ebx,eax
    lodsd
    xor    edx,edx
    lea    edi,.nnx_mx
    call   make_matrixX

    mov    eax,.tc
    mov    ecx,.nxtc
    sub    eax,ecx   ;.nxpc
    lea    eax,[eax*3]
    shl    eax,2
	 ; rotate tris normals
    mov    edi,.tnr  ;[triangles_normals_rotated_ptr]
    add    edi,eax
    mov    esi,edi
    lea    ebx,.nnx_mx
    call   rotary
    ; next points normals
    ; next p. nr. rotated
    mov     edi,.pnr  ;[points_normals_rotated_ptr]
    mov     ecx,.nxpc
    mov     eax,.pc
    sub     eax,.nxpc
    lea     eax,[eax*3]
    shl     eax,2
    add     edi,eax
    mov     esi,edi
    lea     ebx,.nnx_mx
    call   rotary

 .nno_next:
   mov	  esp,ebp
   pop	  ebp
   popad

ret


;=================================================================
rotate_pointsA:
   ;in	ebx - scaled rotary matrix ptr
   push   ebp
   mov	  ebp,esp
   and	  ebp,-16
   sub	  ebp,160
   sub	  esp,320


    .points_count_var	   equ dword[ebp-48]
    .triangles_count_var   equ dword[ebp-44]
    .points_r_ptr	   equ dword[ebp-40]
    .triangles_ptr	   equ dword[ebp-36]
    .points_rotated_ptr    equ dword[ebp-32]
    .pnr		   equ dword[ebp-28]
    .edges_ptr		   equ dword[ebp-24]

    .cur_nodes		   equ dword[ebp-8]
    .tri_no		   equ dword[ebp-4]

    .mid_vert	 equ [ebp-64]
    .mx_ptr	 equ dword[ebp-68]
    .ptr	 equ dword[ebp-72]
    .NxBendDerv  equ dword[ebp-76]
    .bez_cnt	 equ dword[ebp-80]
    .scale3	 equ [ebp-96]
    .zer_h	 equ [ebp-112]

    .scale_brd	 equ [ebp]
    .NxCrCns	 equ [ebp+16]
    .scale_ben	 equ [ebp+32]
    .vert_cnt	 equ [ebp+48]
    .vert_offs	 equ dword[ebp+52]
    .nnx_rot	 equ qword[ebp+56]
    .onedtwo	 equ [ebp+64]
    .chu_ptr	 equ dword[ebp+104]
    .ver_ffd_cnt equ dword[ebp+108]
    .tri_ffd_cnt equ dword[ebp+112]
    .derv_vertNx equ dword[ebp+116] ; 48 bytes
    .nnx_mx	 equ [ebp+120]


    cld
    lea      esi,[points_count_var]
    lea      edi,.points_count_var
    mov      ecx,7
    rep      movsd


    mov      .mx_ptr,ebx	      ; init some locals
    movaps   xmm2,[zero_hgst]
    movaps   .zer_h,xmm2
    xorps    xmm1,xmm1
    movss    xmm0,[scale]
    mov      eax,3
    cvtsi2ss xmm2,eax
    movlps   xmm1,[NextCurvConst]
    shufps   xmm0,xmm0,11000000b
    shufps   xmm2,xmm2,11110011b
    cvtdq2ps xmm1,xmm1
    movaps   .scale_brd,xmm0
    movlps   .NxCrCns,xmm1
    mov      .NxBendDerv,NextBendDerv
    movaps   .scale_ben,xmm2


    cmp     [bezier_flag],0
    je	    .no_bz


    movzx    ecx,[bez_nodes_count]	;*************************************
    mov      esi,[derv_nodes_ptr]	;*************************************
    mov      edi,ffd_nodes_rot		;***Bezier patches work
    mov      ebx,.mx_ptr		;*************************************
    call     rotary			;*************************************

    cvtdq2ps xmm3,[xxadd]
    andps    xmm3,.zer_h
    movzx    ecx,[bez_nodes_count]
    mov      ebx,ffd_nodes_rot
  @@:
    movups   xmm1,[ebx]
    addps    xmm1,xmm3
    movups   [ebx],xmm1
    add      ebx,12
    loop     @b


  .no_tr:

   ; bezier patches verts (and tris)

     movzx   ecx,[bez_patch_count]
     mov     .cur_nodes,ffd_nodes_rot
     mov     edi,.points_rotated_ptr
     mov     ebx,.triangles_ptr
     mov     esi,[derv_rect_ptr]
     xor     edx,edx
     mov    .ver_ffd_cnt,edx	; counters
     mov    .tri_ffd_cnt,edx

     mov    .bez_cnt,ecx
     xor     ecx,ecx
  .Loop_ffd:		  ;  bezier patches
     push    ecx
     mov     ecx,16
     movd    xmm0,edi


     mov     edi,ffd_nodes_unpck
     cld
   @@:
     xor     eax,eax
     lodsw
     imul    eax,12
     add     eax,.cur_nodes ;ffd_nodes_rot
     xchg    esi,eax
     movsd
     movsd
     movsd
     xchg    eax,esi
     loop    @b

     push    esi

     mov     esi,ffd_nodes_unpck
     mov     edi,nodes_aligned
     mov     ecx,3
   .alig:
     push    ecx
     push    esi
     mov     ecx,16
   @@:
     movsd
     lodsd
     lodsd     ;  add  esi,8
     loop    @b
     pop     esi
     pop     ecx
     lodsd     ;  esi + 4
     loop    .alig

     mov     esi,nodes_aligned
     movd    edi,xmm0
     mov     edx, 8 shl 16 + 8
     call    bezier_surface	   ; one patch

     sub     edx,10
     add    .ver_ffd_cnt,ecx
     mov    .tri_ffd_cnt,edx
 ;    sub     edi,12

     mov     eax,ecx
     pop     esi
     pop     ecx

if 0
     push    ebx  ; ebx - cur tri ptr
     imul     eax,ecx	;.bez_cnt
 ;    mov     ebx,.cur_tri_l
 ;     ecx - vertex count, pro one patch
 ;     edx - cur triangles count ( pro one patch )
   @@:
     add    [ebx],eax
     add    [ebx-4],eax
     add    [ebx-8],eax
     sub    ebx,12
     dec    edx
     jnz    @b

     pop    ebx
end if

    ; esi - derives
     inc     ecx

     cmp     ecx,.bez_cnt
  ;   jmp     .end
  ;   jnz     .Loop_ffd 	; last !!

   @@:
     push    .ver_ffd_cnt
     pop     [points_count_var]
     push    .tri_ffd_cnt
     pop     [triangles_count_var]
     jmp     .end

  ;   cmp     [bezier_switched],1
     jmp     .end

;   @@:
;     mov    edi,[points_r_ptr]
;     mov    esi,[points_rotated_ptr]
;     mov    ecx,[points_count_var]
;     lea    ecx,[ecx*3]
;     inc    ecx
;     cld
;     rep    movsd

;     jmp    .end

		    ;****************************
		    ;end Bezier patches fragment
.no_bz: 	    ;****************************

     mov       ebx,.mx_ptr  ;[matrix_scaled]
     mov       ecx,.points_count_var
     mov       esi,.points_r_ptr
     mov       edi,.points_rotated_ptr
     call      rotary

     movzx     eax,[NextScale]
     movzx     ebx,[NextScaleZero]
     sub       eax,ebx
     movd      xmm1,[NextScaleXY]
     movd      xmm2,[NextScaleXYZero]
     pxor      xmm0,xmm0
     punpcklwd xmm1,xmm0
     punpcklwd xmm2,xmm0
     psubd     xmm1,xmm2
     cvtdq2ps  xmm1,xmm1

  ;   xor	eax,eax
     cmp       [NextMed_flag],0
     jne       @f
     cmp       [ffd_flag],0
     je        @f
    ; xm0, xm3 - mid vert	  ; whole ffd
    ; xm6   - scale
     movss     xmm6,[scale]
     xorps     xmm0,xmm0
     xorps     xmm3,xmm3
     shufps    xmm6,xmm6,0
     mov       eax,.points_rotated_ptr
     mov       ecx,.points_count_var
     call      ffd
    @@:


       mov	  ecx,200
       cvtsi2ss   xmm3,ecx
       rcpss	  xmm3,xmm3
       shufps	  xmm3,xmm3,11000000b
       shufps	  xmm4,xmm4,11000000b

       cvtsi2ss   xmm2,eax
       shufps	  xmm2,xmm2,11000000b
       addps	  xmm2,xmm1
       mulps	  xmm2,xmm3
       addps	  xmm2,[the_one]
       movaps	  .scale3,xmm2

       mov	 eax,[NextRotat]
       sub	 eax,dword [NextRotatZero]
       movd	 xmm7,eax
       pxor	 xmm6,xmm6
       punpcklwd xmm7,xmm6
       movq	 .nnx_rot,xmm7

       cld
       lea     esi,.nnx_rot
       lodsd
       xchg    ebx,eax
       lodsd
       xor     edx,edx
       lea     edi,.nnx_mx
       call    make_matrixX
     @@:


 ;---------------------------------
 ;------------ /next/ edition work
 ;---------------------------------

    or	     ebx,-1
    mov      .tri_no,ebx
    mov      al,[NextMed_flag]
    cmp      al,1
    je	     .next
    cmp      al,3
    je	     .whole
    cmp      al,2
    jne      .end
    mov      eax,[curr_chunk]
    cmp      eax,ebx
    je	     .end
    or	     eax,eax
    jz	     .end
    cmp      [chunks_count],1
    je	     .end
    mov      esi,[chunks_desc_ptr]
    cmp      esi,ebx
    je	     .end

; bend
;     mov	  eax,NextBend1        attempt for more precise
;     lea	  edi,.derv_vertNx
;     pxor	  xmm1,xmm1
;     mov	  ecx,4
;   @@:
;     movd	  xmm0,[eax]
;     punpcklwd   xmm0,xmm1
;     cvtdq2ps	  xmm0,xmm0
;     movups	  [edi],xmm0   ; x, y, z = 0
;     add	  edi,12
;     add	  eax,4
;     loop	  @b

;     mov	 eax,NEXT_CURV_NUMB
;     mov	 ebx,1
;     lea	 ecx,.derv_vertNx
;     mov	 edx,NextBendDerv
;     xor	 esi,esi
;     call	 calc_long_curve



    mov      eax,[curr_chunk]
    shl      eax,4
    add      eax,[chunks_desc_ptr]
    cmp      dword[eax+4],0
    jne      @f

    push     eax
    call     opt_chunks       ; calc chunks description struct
    mov      eax,[tex_scale]
    call     calc_bumpmap_coords
    mov      eax,.edges_ptr
    mov      ebx,.triangles_ptr
    mov      ecx,.triangles_count_var
    call     do_edges_list
    mov      [edges_count],ecx
    pop      eax
   @@:
			   ; mov eax,[chunks_desc_ptr]
    mov      ecx,[eax]	   ; [eax]    - tri  count
    mov      edi,[eax+4]   ; [eax+4]  = edi - vert count in curr chun
			   ; [eax+8]  - vert offset
			   ; [eax+12] - tri offset
    mov      edx,dword[eax+8]
    imul     edx,12
    add      edx,.points_rotated_ptr
    jmp      .m_val
   .next:
    cmp     al,1
    jne     .whole
    mov     edi,[NextPointsCount]
    or	    edi,edi
    jz	    .end
    mov     edx,.points_count_var
    sub     edx,edi
    imul    edx,12
    add     edx,.points_rotated_ptr
    jmp     .m_val    ; make values

   .whole:
    cmp      al,3     ; whole object
    jne      @f
    mov      edx,.points_rotated_ptr
    mov      edi,.points_count_var
    mov      ecx,.triangles_count_var
  .m_val:
  @@:
    mov      .vert_offs,edx
    mov      .vert_cnt,edi


    push     edx  ;.vert_offs
    push     edi  ;dword .vert_cnt
    mov      eax,'ocen'      ;	only center translate - find max min also !!
    call     normalize_object1
    ; xm0, xm3 - mid vert
    ; xm6   - scale
    andps    xmm0,.zer_h
    movaps   .mid_vert,xmm0

    xor       eax,eax
    cmp       [ffd_flag],al
    je	      @f
    mov       eax,.vert_offs
    mov       ecx,.vert_cnt
    call      ffd
    jmp       .end
   @@:


    mov      ecx,.vert_cnt
    mov      edi,.vert_offs
    mov      esi,edi
    lea      ebx,.nnx_mx
    call     rotary

    mov      edi,.vert_offs
    xor      ecx,ecx
    mov      ecx,.vert_cnt
    rcpps    xmm7,.scale_brd
    mov      eax,[NextCurvConst]
    lea      eax,[eax*2-1]
    cvtsi2ss xmm6,eax

	; xmm7 - brosdcasted scale  reciprocal
  @@:
    movups   xmm0,[edi]
    mulps    xmm0,xmm7
    movaps   xmm2,xmm0
    mulps    xmm2,.NxCrCns
    addps    xmm2,.NxCrCns
    minps    xmm2,xmm6
    xorps    xmm3,xmm3
    cvtps2dq xmm2,xmm2
    movd     eax,xmm2

    imul     eax,12
    add      eax,.NxBendDerv
    movlps   xmm3,[eax]
    mulps    xmm3,.scale_ben
    addps    xmm0,xmm3

    mulps    xmm0,.scale_brd
    mulps    xmm0,.scale3
    addps    xmm0,.mid_vert
    movlps   [edi],xmm0
    movhlps  xmm0,xmm0
    movss    [edi+8],xmm0
    add      edi,12
    loop     @b

  .chA:


 .end:

;   call   calc_shd_verts

   add	  esp,320
   pop	  ebp

ret


;=======================================================================
ffd:

     ;*****************************
     ;*****************************
     ;Free form modifier
     ;*****************************
     ;*****************************
     ;*****************************

; in:
;   eax -  vertex start offset	; can be one chunk
;   ecx  - vertices count to deform
;

    push	ebp
    mov 	ebp,esp
    sub 	esp,115
    and 	ebp,-16

;in:
;	eax -  start vertices list
;	ecx - vertices count
;	xmm6 - scale
;	xmm3 - center

    .scale_brd	    equ  [ebp-16]
    .onedtwo	    equ  [ebp-32]
    .start_vert     equ  [ebp-36]
    .vert_cnt	    equ  [ebp-40]
    .proc_call	    equ  dword[ebp-44]
    .ffd_flag	    equ  byte[ebp-47]
    .marker	    equ  byte[ebp-48]
    .dest_vert	    equ  [ebp-64]
    .ffd_center     equ  [ebp-80]
    .dest_vert_inti equ  [ebp-96]




    mov      .start_vert,eax
    mov      .vert_cnt,ecx


    movaps   .scale_brd,xmm6
    movaps   [ffd_scale],xmm6
    movaps   [ffd_center],xmm3
    movaps   .ffd_center,xmm3

    movzx    ecx,[ffd_flag]
    mov     .ffd_flag,cl

    mov      ecx,56
    mov      esi,ffd_nodes
    mov      edi,ffd_nodes_rot
    mov      ebx,matrix
    cmp      .ffd_flag,2
    jne      @f
    call     rotary

    mov      ecx,56
    mov      esi,ffd_nodes_intial
    mov      edi,ffd_nodes_intial_rot
    mov      ebx,matrix
    call     rotary
    mov      esi,ffd_nodes_rot
    mov      edi,esi

   @@:
  .ff2:

    movaps   xmm2,.scale_brd

    xor      eax,eax   ; - > MARKER
;    mov      eax,ffd_nodes_rot
    ;ffd_nodes_rot -> 1st dest -  ffd pralellpiped destination
    ;ffd_nodes_intial_rot -> sec destination
    movaps   xmm3,.ffd_center
    movaps   xmm4,[zero_hgst]
    andps    xmm3,xmm4
    andps    xmm2,xmm4
    orps     xmm2,[one_hgst_dd]

  .ag1:
    mov      ecx,56
  @@:
    movups   xmm0,[esi]
    mulps    xmm0,xmm2
    addps    xmm0,xmm3
    movups   [edi],xmm0
    add      esi,12
    add      edi,12
    loop     @b
    mov      esi,ffd_nodes_intial_rot
    mov      edi,esi
    inc      eax
    cmp      eax,1
    je	     .ag1




     mov    eax,64
     mov    ecx,16
     cmp    .ffd_flag,2
     cmove  ecx,eax
     xor    edx,edx	  ; marker
     mov    .marker,dl
     mov    edx,ffd_nodes_rot
     mov    esi,ffd_rect
     mov    edi,ffd_nodes_unpck
     cld
   @@:
     xor    eax,eax
     lodsw
     imul   eax,12
     add    eax,edx
     xchg   esi,eax
     movsd
     movsd
     movsd
     xchg   eax,esi
     loop   @b
     mov    esi,ffd_rect
     mov    edi,ffd_nodes_unpck_inti
     mov    edx,ffd_nodes_intial_rot
     mov    ecx,64
     inc    .marker
     cmp    .marker,1
     je      @b


     mov    ebx,4
     cmp    .ffd_flag,1
     mov    edx,1
     cmove  ebx,edx
     xor    eax,eax
     mov    .marker,al

     mov    esi,ffd_nodes_unpck
     mov    edi,nodes_aligned	 ; minimum 16*16*3 bytes - in Bezier ffd volume cause
   .alig22:
     mov    ecx,3
   .alig2:
     push   ecx
     push   esi
     mov    ecx,16
     imul   ecx,ebx
   @@:
     movsd
     lodsd
     lodsd     ;  add  esi,8
     loop   @b
     pop    esi
     pop    ecx
     lodsd     ;  esi + 4
     loop   .alig2


     mov    esi,ffd_nodes_unpck_inti
     mov    edi,nodes_aligned_inti    ; minimum 16*16*3 byt
     inc    .marker
     cmp    .marker,1
     je     .alig22



     mov     ebx,1
     mov     eax,bezier_surface
     mov     ebx,bezier_volume
     cmp     .ffd_flag,2
     cmove   eax,ebx
     mov     .proc_call,eax


   mov	     ebx,.start_vert  ;[points_rotated_ptr]
   mov	     ecx,.vert_cnt    ;[points_count_var]
   mov	     eax,2
   cvtsi2ss  xmm1,eax
   shufps    xmm1,xmm1,0
   rcpps     xmm1,xmm1
   movaps    .onedtwo,xmm1


 .llopA:		    ; free form modifier
   push      ecx
   movups    xmm0,[ebx]
   divps     xmm0,.scale_brd
   mulps     xmm0,.onedtwo
   addps     xmm0,.onedtwo
   ;xm0 0.0 - 1.0
   movaps    .dest_vert,xmm0
   movaps    xmm1,xmm0
   movhlps   xmm2,xmm0
   shufps    xmm1,xmm1,11111101b

    push      ebx
    cmp       .ffd_flag,2
    jne       @f
    mov       esi,nodes_aligned_inti	; ffd_verts_rot
    lea       edi,.dest_vert_inti
    call      .proc_call  ;bezier_surface
  @@:
   movaps    xmm0,.dest_vert
   movaps    xmm1,xmm0
   movhlps   xmm2,xmm0
   shufps    xmm1,xmm1,11111101b
   mov	     esi,nodes_aligned	  ; ffd_verts_rot
   mov	     eax,'offd' 	  ; only one vert
   lea	     edi,.dest_vert
   call      .proc_call  ;bezier_surface
   movaps    xmm0,.dest_vert
   pop	     ebx
   cmp	     .ffd_flag,2
   je	     @f
   movlps    [ebx],xmm0
   jmp	     .ff1
  @@:
   movaps    xmm1,.dest_vert_inti
   subps     xmm0,xmm1
   movups    xmm3,[ebx]
   addps     xmm3,xmm0
   movlps    [ebx],xmm3
   movhlps   xmm3,xmm3
   movss     [ebx+8],xmm3
  .ff1:
  @@:


   add	     ebx,12
   pop	     ecx
   loop      .llopA

 .no_ffd_modify:

   add	esp,115
   pop	ebp
ret
;=======================================================

;==========================================
peel:
   pushad
   push  ebp
   mov	 ebp,esp
   sub	 esp,128+80
   and	 ebp,-16
   sub	 ebp,64


    .x_max		   equ dword [ebp-4]
    .x_min		   equ dword [ebp-8]
    .y_max		   equ dword [ebp-12]
    .y_min		   equ	     [ebp-16]
    .th_no		   equ dword [ebp-20]
    .rplc		   equ dword [ebp-24]	   ;  real_phg_line_col
    .bilf		   equ dword [ebp-28]	   ;  r_phg_bf_line_z
    .glas		   equ dword [ebp-32]	   ;  glass_line
    .gltl		   equ	     [ebp-36]	   ;  glass_tex_line
    .gtsl		   equ	     [ebp-40]	   ;  glass_tex_line_sl
    .gtsv		   equ	     [ebp-44]	   ;  glass_tex_line_sl_var
    .furl		   equ	     [ebp-48]	   ;  fur_real_phong_line_z
    .shdl		   equ dword [ebp-52]	   ;  ray_shd_l
    .sli_var		   equ dword [ebp-56]	   ;  st_sl_var_line

    .coord3		   equ [ebp-84]
    .coord2		   equ [ebp-88]
    .coord1		   equ [ebp-92]
    .esi_pee_call	   equ dword[ebp-96]
    .line_call		   equ dword[ebp-100]
    .dr_fl		   equ byte [ebp-101]

    .points_count_var	   equ dword[ebp]
    .triangles_count_var   equ dword[ebp+4]
    .points_r_ptr	   equ dword[ebp+8]
    .triangles_ptr	   equ dword[ebp+12]
    .points_rot 	   equ dword[ebp+16]
    .pnr		   equ dword[ebp+20]
    .edges_ptr		   equ dword[ebp+24]
    .edges_count	   equ dword[ebp+28]
    .edge_s_d_ptr	   equ dword[ebp+32]
    .screen_ptr 	   equ dword[ebp+36]
    .Zbuffer_ptr	   equ dword[ebp+40]
    .slices_counter_ptr    equ dword[ebp+44]
    .slices_ptrs_buff_ptr  equ dword[ebp+48]
    .tex_points_f_ptr	   equ dword[ebp+52]
    .zer_h		   equ [ebp+64]


    cld
    lea      esi,[sli_var]    ; copying some constans -
    lea      edi,.sli_var     ; horiontal line adresses
    mov      ecx,9	      ; rendering procs
    rep      movsd

    lea      esi,[points_count_var]
    lea      edi,.points_count_var
    mov      ecx,14
    rep      movsd

    mov     .th_no,ebx
    mov     edx,.slices_ptrs_buff_ptr
    emms

    mov     eax,.gtsv
    mov     esi,edx
    movzx   ebx,[draw_flag]
    mov     .dr_fl,bl
    cmp     ebx,19
    cmove   eax,.gtsl
    cmove   esi,.Zbuffer_ptr
    cmp     ebx,18
    cmove   eax,.gltl
    cmp     ebx,17
    cmove   eax,.glas
    mov     .line_call,eax
    cmp     ebx,16
    cmove   eax,.bilf
    cmove   esi,.Zbuffer_ptr
    cmp     ebx,15
    cmove   eax,.rplc
    cmp     ebx,21
    cmove   eax,.furl
    cmove   esi,.Zbuffer_ptr
    cmp     ebx,100
    cmove   eax,.shdl
    cmove   esi,.slices_ptrs_buff_ptr

    mov     .line_call,eax
    mov     .esi_pee_call,esi ; zbuff / slices ptr buff

    movaps  xmm0,[zero_hgst]
    movaps  .zer_h,xmm0
    movzx   ebx,[yres_var]
    mov     eax,ebx
    shr     ebx,1
    movzx   edx,[xres_var]
    movzx   ecx,dx
    shr     edx,1
    xor     esi,esi

    cmp    .th_no,esi
    jne    @f
    mov    .y_min,esi
    mov    .x_min,esi
    mov    .y_max,ebx
    mov    .x_max,edx
    jmp    .skip_div
   @@:
    cmp    .th_no,1
    jne    @f
    mov    .y_min,esi
    mov    .x_min,edx
    mov    .x_max,ecx
    mov    .y_max,ebx
    jmp    .skip_div
   @@:
    cmp    .th_no,2
    jne    @f
    mov    .y_max,eax
    mov    .y_min,ebx
    mov    .x_min,esi
    mov    .x_max,edx
    jmp    .skip_div
  @@:
    cmp    .th_no,3
    jne    @f

    mov    .y_max,eax
    mov    .y_min,ebx
    mov    .x_min,edx
    mov    .x_max,ecx

  @@:
  .skip_div:
    cmp       [shadow_flag],1
    jne       @f
    movups    xmm1,.y_min
  ;  xorps     xmm2,xmm2
    call      do_shadow
  @@:
    cmp    .dr_fl,20
    jne    .pee

    mov    esi,.triangles_ptr
    mov    ecx,.triangles_count_var
    mov    edi,.slices_ptrs_buff_ptr
    mov    edx,.slices_counter_ptr

  .again_stencil_tri:

    push   ecx
    lodsd
    xchg   ecx,eax
    lodsd
    xchg   ebx,eax    ; shorter than mov  ebx,eax
    lodsd
    xchg   ecx,eax
    push   esi

    imul   eax,12
    imul   ebx,12
    imul   ecx,12
    add    eax,.points_rot
    add    ebx,.points_rot
    add    ecx,.points_rot

    push     eax dword[ecx+8] dword[ebx+8] dword[eax+8]
    movups   xmm3,[esp] ; xmm3 - z float pack

    movlps   xmm0,[eax]
    movhps   xmm0,[ebx]
    movups   xmm2,[ecx]
    cvtps2dq xmm0,xmm0
    cvtps2dq xmm2,xmm2
    packssdw xmm0,xmm0
    packssdw xmm2,xmm2
    movlhps  xmm0,xmm2
    movups   [esp],xmm0
    pop      eax ebx ecx
    add      esp,4
    ror      eax,16
    ror      ebx,16
    ror      ecx,16
    movaps   xmm0,xmm3
    movaps   xmm1,.y_min
    movaps   xmm0,xmm3
    mov      esi,.slices_ptrs_buff_ptr
    movd     xmm5,.sli_var
    push     ebp
    call     stencil_tri
    pop      ebp
    pop      esi
    pop      ecx
 ;   dec      ecx
 ;   jnz      .again_stencil_tri
    loop     .again_stencil_tri

   .pee:
    cld
    mov      esi,.triangles_ptr
    xor      ecx,ecx
    mov      edi,.screen_ptr
    cld
  .again_peel_tri:
    push     ecx
    movd     mm7,ecx
    push     edi

    lodsd
    mov      ecx,eax
    lodsd
    mov      edx,eax
    lodsd
    xchg     ecx,eax
    push     eax
    push     edx
    push     ecx
    imul     eax,12
    imul     edx,12
    imul     ecx,12

    mov      ebx,.pnr
    movups   xmm7,.zer_h
    movups   xmm0,[eax+ebx]
    movups   xmm1,[edx+ebx]
    movups   xmm2,[ecx+ebx]
    andps    xmm0,xmm7
    andps    xmm1,xmm7
    andps    xmm2,xmm7
    mov      ebx,.points_rot  ;.points_rot
    movlps   xmm4,[eax+ebx]
    movhps   xmm4,[edx+ebx]
    movups   xmm6,[ecx+ebx]

    movhlps  xmm7,xmm6
    cvtps2dq xmm4,xmm4
    cvtps2dq xmm6,xmm6
    packssdw xmm4,xmm4
    packssdw xmm6,xmm6
    movlhps  xmm4,xmm6

    movups   .coord1,xmm4
    movlps   xmm4,[eax+ebx+8]
    movhps   xmm4,[edx+ebx+8]
    shufps   xmm4,xmm4,11111000b  ; z coord work
    movlhps  xmm4,xmm7
    movups   xmm5,.y_min
    pop      ecx
    pop      edx
    pop      eax
    shl      eax,3
    shl      edx,3
    shl      ecx,3
    mov      ebx,.tex_points_f_ptr
    movlps   xmm6,[eax+ebx]
    movhps   xmm6,[edx+ebx]
    movlps   xmm3,[ecx+ebx]

    push     esi
    mov      eax,0x00aa3333
    movd     mm0,eax	 ; r phg col
    mov      eax,.coord1
    mov      ebx,.coord2
    mov      ecx,.coord3
    ror      eax,16
    ror      ebx,16
    ror      ecx,16

    mov      edx,texmap
    mov      esi,.esi_pee_call	;.slc_ptrs
    cmp      .dr_fl,17
    cmove    edx,esi

    movd     xmm7,.line_call
    push     ebp	   ; my core duo xp needs this push to properly work..
    call     glass_tex_tri
    pop      ebp
    pop      esi
    pop      edi
    pop      ecx
    inc      ecx
    cmp      ecx,.triangles_count_var
    jne      .again_peel_tri

    add      esp,128+80
    pop      ebp
    popad
ret
;=======================================================
;=================================================================

translate_points:
    pushad
    push     ebp
    mov      ebp,esp
    sub      esp,60

    .points_count_var	   equ dword[ebp-52]
    .triangles_count_var   equ dword[ebp-48]
    .points_r_ptr	   equ dword[ebp-44]
    .triangles_ptr	   equ dword[ebp-40]
    .points_rotated_ptr    equ dword[ebp-36]
    .pnr		   equ dword[ebp-32]
    .edges_ptr		   equ dword[ebp-28]
    .edges_count	   equ dword[ebp-24]
    .edge_s_d_ptr	   equ dword[ebp-20]
    .screen_ptr 	   equ dword[ebp-16]
    .Zbuffer_ptr	   equ dword[ebp-12]
    .ptr		   equ [ebp-4]

    cld
    lea      esi,[points_count_var]
    lea      edi,.points_count_var
    mov      ecx,11
    rep      movsd


    cmp      [bezier_flag],0
    jne       .end2


    xor      esi,esi
    cvtdq2ps xmm3,[xxadd]
    mov      ebx,.points_rotated_ptr
    mov      ecx,.points_count_var
    sub      ecx,[NextPointsCount]
    andps    xmm3,[zero_hgst]
  .add1:
    movups   xmm1,[ebx]
    addps    xmm1,xmm3
    movups   [ebx],xmm1
    add      ebx,12
    loop     .add1

    cmp      [ffd_flag],0
    je	     .no_ffd

    cmp      esi,0
    cmove    edi,ebx
;    cmp      [ffd_flag],1
;    jne      @f
    mov      ebx,ffd_nodes_rot
    mov      ecx,56
    inc      esi
    cmp      esi,2
    jne      .add1

;    inc      esi
;    mov      ebx,ffd_verts_rot
;    mov      ecx,[ffd_vert_cnt]  ;56 ;.ver_ffd_cnt
;    mov      eax,ffd_nodes_rot
;    mov      edx,56
;    cmp      esi,2
;    cmove    ebx,eax	;ffd_verts_rot
;    cmove    ecx,edx  ;.ver_ffd_cnt
;    cmp      esi,3
;    jne      .add1
    xchg      ebx,edi
;  @@:


  .no_ffd:

    movq      xmm7,[NextMxadd]
    movq      xmm5,[NextMsub]
    pxor      xmm6,xmm6
    punpcklwd xmm7,xmm6
    punpcklwd xmm5,xmm6
    psubd     xmm7,xmm5
    pslld     xmm7,1
    cvtdq2ps  xmm7,xmm7

    mov       ecx,[NextPointsCount]
    jecxz     .end_nx
    addps     xmm3,xmm7
    andps     xmm3,[zero_hgst]
  @@:
    movups   xmm1,[ebx]
    addps    xmm1,xmm3
    movups   [ebx],xmm1
    add      ebx,12
    loop     @b
    jmp      .end2

   .end_nx:
    cmp      [NextMed_flag],2
    jne      .end2
    cmp      [curr_chunk],-1
    je	     .end2
    andps    xmm7,[zero_hgst]

    mov      esi,[long_pipe_vert_ptr]
    mov      ecx,4
  @@:
    movups   xmm0,[esi]
    addps    xmm0,xmm3
    movups   [esi],xmm0
    add      esi,12
    loop     @b



    mov      ebx,.triangles_count_var
    add      ebx,200
    malloc   ebx
    mov      .ptr,eax


    mov      edi,eax
    cld
    xor      eax,eax
    mov      ecx,ebx
    sub      ecx,10
    shr      ecx,2
    rep      stosd

    mov       ebx,.triangles_ptr
    mov       ecx,.triangles_count_var
    mov       esi,[chunks_ptr]
    cld
    mov       edx,[curr_chunk]
    andps     xmm7,[zero_hgst]

   .lab1:
    push      ecx
    lodsw
    xor       ecx,ecx
   .lab2:
    push      eax
    cmp       ax,dx
    jne       @f
    mov       eax,[ebx+ecx*4]
    mov       edi,eax
    add       edi,.ptr
    cmp       byte[edi],1
    je	      @f
    imul      eax,12
    add       eax,.points_rotated_ptr
    movups    xmm1,[eax]
    addps     xmm1,xmm7
    movups    [eax],xmm1
    mov       byte[edi],1
   @@:
    pop       eax
    inc       ecx
    cmp       ecx,3
    jnz       .lab2
    pop       ecx
    add       ebx,12
    loop      .lab1


    invoke    VirtualFree ,.ptr,0 , MEM_RELEASE

   .end2:
;    call      calc_shd_verts

    mov       esp,ebp
    pop       ebp
    popad
ret

;==============================================================
if 0
calc_shd_verts:
       push	 ebx esi edi
       push	 ebp
       mov	 ebp,esp
       sub	 esp,200
       and	 ebp,-16
       sub	 ebp,112
       .p_light_vec  equ [ebp]	      ; one light directional vector
       .shd_mx	     equ [ebp+32]
       .counter      equ [ebp+100]
       .shd_vertices equ [ebp-48]
       .zer_h	     equ [ebp-64]

	lea	ebx,.shd_mx	       ; 64 bytes
	lea	esi,[lights_aligned]   ;.p_light_vec	   ; +eax
	call	calc_shadow	       ; calculate 4x4 shadow matrix

     .next_shd_vert:
	mov	esi,[points_rotated_ptr]
	mov	ecx,[points_count_var]
	lea	ebx,.shd_mx
	lea	edi,[shadow_vertices_ptr]  ; 16  bytes per one vertex
	add	edi,eax
	call	mul_matrix4x4

	; thats all ??

	add	esp,200
	mov	ebp,esp
	pop	ebp
	pop	edi esi ebx
ret
end if
;============================================================================
;============================================================================
do_shadow:
; in:	    xmm1 -> ymin ymax xmin xmax
;
       pushad
       push	 ebp
       mov	 ebp,esp
       sub	 esp,250
       and	 ebp,-16
       sub	 ebp,128

       .p_light_vec  equ [ebp]
       .shd_mx	     equ [ebp+32]
       .counter      equ [ebp+100]
       .shd_vertices equ [ebp-48]
       .zer_h	     equ [ebp-64]
       .xmax	     equ [ebp-68]
       .xmin	     equ [ebp-72]
       .ymax	     equ [ebp-76]
       .ymin	     equ [ebp-80]
       .ptsr	      equ dword[ebp-84]
       .pl_coords     equ dword[ebp-88]
       .mul_matrix4x4 equ dword[ebp-92]
       .calc_shadow   equ dword[ebp-96]
       .scr	      equ dword[ebp-100]
       .shd_buffer    equ dword[ebp-104]

	push	[shd_buffer_ptr]
	push	[screen_ptr]
	pop	.scr
	pop	.shd_buffer
	push	[points_rotated_ptr]
	pop	.ptsr
	mov	.pl_coords,ppoint_light_coords
	mov	.mul_matrix4x4,mul_matrix4x4
	mov	.calc_shadow,calc_shadow
	movaps	xmm0,[zero_hgst_dd]
	movaps	.ymin,xmm1
	movaps	.zer_h,xmm0   ; some definitions to decrease loop size...

	mov	ecx,[triangles_count_var]
	mov	esi,[triangles_ptr]
   .again_shd_tri:
	push	esi
	push	ecx
	xor	ecx,ecx
     .next_shd_vert:
	push	esi
	mov	.counter,ecx
	mov	eax,ecx
	shl	eax,2
	mov	edx,dword[eax+esi]
	imul	edx,12
	add	edx,.ptsr   ; [edx] - cur vertex
	movups	xmm0,[edx]
	push	edx
	mov	eax,.pl_coords
	andps	xmm0,.zer_h
	subps	xmm0,[eax]
	movaps	xmm1,xmm0
	mulps	xmm1,xmm1
	haddps	xmm1,xmm1
	haddps	xmm1,xmm1
	andps	xmm1,.zer_h
	rsqrtps xmm1,xmm1
	mulps	xmm0,xmm1	   ; xmm0 - normalized light vector
	andps	xmm0,.zer_h
	movups	.p_light_vec,xmm0  ;+eax

	lea	ebx,.shd_mx	       ; 64 bytes
	lea	esi,.p_light_vec       ; +eax
	call	.calc_shadow		; calculate 4x4 shadow matrix

	pop	esi
	mov	ecx,1
	mov	eax,.counter
	shl	eax,4
	lea	ebx,.shd_mx
	lea	edi,.shd_vertices  ; 16  bytes per one vertex
	add	edi,eax
	call	.mul_matrix4x4

	pop	esi
	mov	ecx,.counter
	inc	ecx
	cmp	ecx,3
	jnz	.next_shd_vert

	lea	 ebx,.shd_vertices
	movlps	 xmm0,[ebx]
	movhps	 xmm0,[ebx+16]
	movups	 xmm1,[ebx+32]
	cvtps2dq xmm0,xmm0
	cvtps2dq xmm1,xmm1
	movdqa	 xmm3,xmm0
	movhlps  xmm4,xmm0
	movdqa	 xmm5,xmm1
	movdqa	 xmm6,.ymin
	pshufd	 xmm6,xmm6,01110010b ; l->h xmin, ymin, xmax, ymax
	movlhps  xmm3,xmm3
	movlhps  xmm4,xmm4
	movlhps  xmm5,xmm5

	pcmpgtd  xmm3,xmm6
	pcmpgtd  xmm4,xmm6
	pcmpgtd  xmm5,xmm6

	movhlps  xmm6,xmm3
	pxor	 xmm3,xmm6
	movhlps  xmm6,xmm4
	pxor	 xmm4,xmm6
	movhlps  xmm6,xmm5
	pxor	 xmm6,xmm5

	por	 xmm3,xmm4
	por	 xmm3,xmm6
	pmovmskb eax,xmm3
	or	 al,al
	jz	 .again
	packssdw xmm0,xmm0	 ; xm0	lo -> 2 words
	packssdw xmm1,xmm1	 ; xm2	lo -> 4 words
	movlhps  xmm0,xmm1
	sub	 esp,16
	movups	 [esp],xmm0
	pop	 ecx ebx eax edx
	ror	 eax,16
	ror	 ebx,16
	ror	 ecx,16
	mov	 edx,[xres_vard]
	ror	 edx,16
	mov	 edi,.scr
	mov	 esi,.shd_buffer
	movaps	 xmm1,.ymin
	movd	 xmm5,[shal]  ; shadow line
	call	 stencil_tri  ; triangle_shd
     .again:
	pop	  ecx
	pop	  esi
	add	  esi,12
	dec	  ecx
	jnz	 .again_shd_tri

	add	 esp,250
	pop	 ebp
	popad
ret
;=============================================================
if 0
do_stencil:
    push	ebp
    mov 	ebp,esp
    sub 	esp,30
;in :	  xmm5 - miny ymax xmin xmax
       .stencil_esi equ dword [ebp-4]
       .stencil_hor equ dword [ebp-8]
       .xmax	     equ [ebp-16]
       .xmin	     equ [ebp-20]
       .ymax	     equ [ebp-24]
       .ymin	     equ [ebp-28]

    movups   .ymin,xmm5
    mov      al,[draw_flag]
    mov      esi,[slices_ptrs_buff_ptr]  ;.sl_ptrs_b_ptr
    mov      edi,[shll]
    cmp      al,20
    cmove    esi,[slices_counter_ptr]  ;.slc_cnt_ptr
    cmove    edi,[cssl]
    cmp      al,19
    cmove    esi,[Zbuffer_ptr]
    cmove    edi,[sshl]
    cmp      al,100
    cmove    esi,[slices_ptrs_buff_ptr]
    mov      .stencil_esi,esi
    mov      .stencil_hor,edi
    mov      esi,[triangles_ptr]
    mov      ecx,[triangles_count_var]
    mov      edx,[points_rotated_ptr]
  .again_stencil_tri:
  @@:
    push   edx
    push   ecx
    lodsd
    xchg   eax,ecx
    lodsd
    xchg   ebx,eax
    lodsd
    xchg   eax,ecx
    push   esi
    imul   eax,12
    imul   ebx,12
    imul   ecx,12
  ;  add    eax,[points_rotated_ptr]
  ;  add    ebx,[points_rotated_ptr]
  ;  add    ecx,[points_rotated_ptr]
    movss  xmm3,[edx+eax+8]
    movss  xmm4,[edx+ebx+8]
    movss  xmm5,[edx+ecx+8]
    movups xmm0,[eax+edx]
    movups xmm1,[ebx+edx]
    movups xmm2,[ecx+edx]
    movss  xmm3,[edx+eax+8]
    xorps   xmm4,xmm4
;    movhlps xmm3,xmm0
    movhlps xmm4,xmm1
    movhlps xmm5,xmm2

    shufps  xmm4,xmm4,11110011b

    orps    xmm3,xmm4

    movlhps  xmm3,xmm5	; xmm3 - z float pack
    cvtps2dq xmm0,xmm0
    cvtps2dq xmm1,xmm1
    cvtps2dq xmm2,xmm2
    packssdw xmm0,xmm0
    packssdw xmm1,xmm1
    packssdw xmm2,xmm2
    movd     eax,xmm0
    movd     ebx,xmm1
    movd     ecx,xmm2
    ror      eax,16
    ror      ebx,16
    ror      ecx,16
    movups   xmm1,.ymin
    movaps   xmm0,xmm3
    mov      edx,[xres_vard]
    ror      edx,16
    mov      esi,.stencil_esi
    movd     xmm5,.stencil_hor
    call     stencil_tri
    pop      esi
    pop      ecx
    pop      edx
    dec      ecx
    jnz      .again_stencil_tri

    mov     esp,ebp
    pop     ebp
ret
end if
;==================================================================
;==================================================================
;==================================================================
;==================BEGIN MAIN RENDERING PROC=========================
;==================================================================
;==================================================================
;==================================================================

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; g_draw - draw to screen buffer; main program proc
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
g_draw:
    push ebx esi edi

    push	ebp
    mov 	ebp,esp
    sub 	esp,128


    .xres_var	    equ     word [ebp-4]
    .yres_var	    equ     word [ebp-2]
    .xres_vard	    equ     dword[ebp-4]
    .scr	    equ     dword[ebp-8]       ;  \
    .zbuff	    equ     dword[ebp-12]      ;  |  > dont changr
    .tex_ptr	    equ     dword[ebp-16]      ;  |    order !!
    .width	    equ     [ebp-20]	       ; /
    .dr_fl	    equ     byte [ebp-24 ]
    .slc_cnt_ptr    equ     dword[ebp-28]
    .sl_ptrs_b_ptr  equ     dword[ebp-32]
    .mx 	    equ     [ebp-70]
    .ffd_verts_rot  equ     dword[ebp-74]
    .inner_vert_ptr equ     dword[ebp-78]
    .points_count_var equ   dword[ebp-82]
    .stencil_hor    equ     [ebp-86]
    .stencil_esi    equ     [ebp-90]
    .matrix_scaled  equ     dword[ebp-94]
    .edges_ptr	    equ     dword[ebp-98]
    .points_rotated_ptr equ dword[ebp-102]
    .col	    equ     dword[ebp-106]
    .x_max		   equ dword [ebp-110]
    .x_min		   equ dword [ebp-114]
    .y_max		   equ dword [ebp-118]
    .y_min		   equ	     [ebp-122]


  @@:
    mov     eax,[xres_vard]
    mov    .xres_vard,eax
  ;  mov    eax,.xres_vard
    mov    esi,eax
    cwde
    shr    esi,16


;	movd	  xmm1,[xres_vard]
;	xorps	  xmm3,xmm3
;	punpcklwd xmm1,xmm3
;	pshufd	  xmm1,xmm1,00110111b
;	movdqu	  .y_min,xmm1


    mov    .x_max,eax
    mov    .y_max,esi
    mov    .width,eax
    xor    eax,eax
    mov    .x_min,eax
    mov    .y_min,eax



    movzx  eax,.xres_var
    mov    .width,eax
    push   [screen_ptr]
    pop    .scr
    push   [points_rotated_ptr]
    pop    .points_rotated_ptr
    mov    al,[draw_flag]
    mov    .dr_fl,al
    push   [Zbuffer_ptr]
    pop    .zbuff
    push   [slices_counter_ptr]; [points_count_var]
    pop    .slc_cnt_ptr  ;.points_count_var
    push   [slices_ptrs_buff_ptr]
    pop    .sl_ptrs_b_ptr
    mov    .ffd_verts_rot,ffd_verts_rot
    push   [inner_vert_ptr]
    pop    .inner_vert_ptr
    push   [points_count_var]
    pop    .points_count_var
    mov    .matrix_scaled,matrix_scaled
    push   [edges_ptr]
    pop    .edges_ptr
    cmp    [blur_flag],1
    jne    .bl

    xor    eax,eax
    cmp    .dr_fl,21
    jl	   @f
    or	   eax,-1	   ; clear draw buffer
  @@:
    mov    edi,.scr
    movzx  ebx,.xres_var
    movzx  ecx,.yres_var
    imul   ecx,ebx
    cld
    rep    stosd
    jmp    .no_blur
  @@:
  .bl:
    mov   ecx,1
    call  blur_screen
  .no_blur:
    mov   al,.dr_fl
    cmp   al,1
    je	  @f
    cmp   al,9
    jge   @f
    cmp   al,13
    jge   @f
    mov   eax,5000 shl	ROUND	; make background in Zbuffer
    jmp   .m_back
  @@:
    mov   eax,60000.1
  .m_back:
    mov   edi,.zbuff
    movzx ecx,.xres_var
    movzx ebx,.yres_var
    imul  ecx,ebx
    cmp   .dr_fl,19
    jne   @f
    shl   ecx,2  ; slices need more memory
   @@:
    cld
    rep   stosd

    cmp   .dr_fl,15
    ; all real phong models
    jl	  .rotate
  @@:
    mov   eax,60000.1
    mov   edi,.sl_ptrs_b_ptr
 ;  ; i use temporaly this buff as stencil buff
    movzx ecx,.xres_var
    movzx ebx,.yres_var
    imul  ecx,ebx
    cld
    rep   stosd

  .rotate:
    mov   esi,main_rotary
    mov   eax,[esi+8]
    mov   ebx,[esi]
    mov   edx,[esi+4]
    mov   edi,matrix
    mov   esi,one_deg

    call  make_matrixX

    mov   esi,matrix
    mov   edi,.matrix_scaled
    mov   ecx,36/4
    cld
    rep   movsd

    mov   esi,.matrix_scaled  ;esi - pointer to 3x3 matrix
    mov   ebx,scale	      ;ebx - ptr to scale

    call  add_scale_to_matrix
    mov   ebx,.matrix_scaled

    invoke CreateThread,NULL,NULL,rotate_pointsA,\
	   ebx,NORMAL_PRIORITY_CLASS,\
	   [ThreadID0]

    mov    [hthread0],eax
    invoke SetThreadPriority,hthread0,NORMAL_PRIORITY_CLASS
    invoke SetThreadAffinityMask,[hthread0],1

    invoke CreateThread,NULL,NULL,rotate_normals,\
	   NULL,NORMAL_PRIORITY_CLASS,\
	   [ThreadID1]
    mov    [hthread1],eax
    invoke SetThreadPriority,hthread1,NORMAL_PRIORITY_CLASS
    invoke SetThreadAffinityMask,[hthread1],2

    invoke ResumeThread,[hthread0]
    invoke ResumeThread,[hthread1]

    invoke WaitForMultipleObjectsEx,2,hthread0,1,-1,0
    invoke CloseHandle,[hthread0]
    invoke CloseHandle,[hthread1]
      ;make background in shd buff
	movzx eax,.xres_var
	movzx ecx,.yres_var
	mul   ecx
	mov   ecx,eax
	shr   ecx,2
	inc   ecx
	xor   eax,eax
	mov   edi,[shd_buffer_ptr]   ; 1 byte
	cld
	rep   stosd

;	 call  calc_shd_verts
   cmp	[shadow_flag],1
   jne	.no_shadow


;   cmp   .dr_fl,15
;      .no_shadow
  ;;  jnz    .no_shadow

     ;make background in shd buff

     ;	 movzx eax,.xres_var
     ;	 movzx ecx,.yres_var
     ;	 mul   ecx
     ;	 mov   ecx,eax
     ;	 shr   ecx,2
     ;	 inc   ecx
     ;	 xor   eax,eax
     ;	 mov   edi,[shd_buffer_ptr]   ; 1 byte
     ;	 cld
     ;	 rep   stosd

    ;  draw background image

	cld
	mov	  edi,.scr  ;[screen_ptr]
	mov	  esi,texmap
	xor	  ecx,ecx
     .agg:
	push	  ecx
	xor	  edx,edx
     @@:
	push	  edx
	and	  edx,TEX_X - 1  ;0x1ff
	and	  ecx,TEX_Y - 1 ; 0x1ff
	push	  ecx
	shl	  ecx,TEX_SHIFT
   ;	 imul	   ecx,ebx
	add	  ecx,edx
	shl	  ecx,2
	lea	  esi,[ecx+texmap]
	movsd
	pop	  ecx
	pop	  edx
	inc	  edx
	cmp	  dx,.xres_var
	jnz	  @b

	pop	  ecx
	inc	  ecx
	cmp	  cx,.yres_var
	jnz	  .agg
	cmp	  .dr_fl,15
	jge	   @f

  ;	 movups     xmm1,.y_min
  ;	 cvtdq2ps   xmm2,[xxadd]
  ;	 call	    do_shadow
       @@:
  .no_shadow:
    or	       ebx,-1
    mov        eax,[edit_particle_no]
    cmp        eax,ebx
    je	       .no_editt
    movd       xmm0,dword[edit_end_x]
    movd       xmm1,dword[edit_start_x]
    pxor       xmm2,xmm2
    punpcklwd  xmm0,xmm2
    punpcklwd  xmm1,xmm2
    psubd      xmm0,xmm1
    cvtdq2ps   xmm0,xmm0

;    mov	eax,[edit_particle_no]
    movzx      ebx,[edit_flag]
    cmp        ebx,1
    jne        @f
    dec        eax
    imul       eax,12
    add        eax,.points_rotated_ptr
    movlps     xmm1,[eax]
    addps      xmm1,xmm0
    movlps     [eax],xmm1
    jmp        .no_editt

  @@:	  ; face temporaly edition routines
    cmp        ebx,2
    jnz        @f
    imul       eax,12
    add        eax,[triangles_ptr]
    mov        ebx,dword[eax+4]
    mov        ecx,dword[eax+8]
    mov        eax,dword[eax]
    imul       eax,12
    imul       ebx,12
    imul       ecx,12
    add        eax,.points_rotated_ptr
    add        ebx,.points_rotated_ptr
    add        ecx,.points_rotated_ptr
    movlps     xmm1,[eax]
    movlps     xmm2,[ebx]
    movlps     xmm3,[ecx]
    addps      xmm1,xmm0
    addps      xmm2,xmm0
    addps      xmm3,xmm0
    movlps     [eax],xmm1
    movlps     [ebx],xmm2
    movlps     [ecx],xmm3
    jmp        .no_editt
  @@:
    cmp        ebx,3  ;  dword[edit_edge_no],-1
    jne        @f
    dec        eax
    shl        eax,3
    add        eax,.edges_ptr
    mov        ebx,[eax+4]
    mov        eax,[eax]
    imul       eax,12
    imul       ebx,12
    add        eax,.points_rotated_ptr
    add        ebx,.points_rotated_ptr

    movlps     xmm1,[eax]
    movlps     xmm2,[ebx]
    addps      xmm1,xmm0
    addps      xmm2,xmm0
    movlps     [eax],xmm1
    movlps     [ebx],xmm2

  .no_editt:
  @@:			  ; translate vertices into 2d
     call   translate_points




     cmp    .dr_fl,15
     jnge   .no_peel
     xor    eax,eax	     ; clear  buffer
     mov    edi,.slc_cnt_ptr
     movzx  ebx,.xres_var
     movzx  ecx,.yres_var
     imul   ecx,ebx
     shr    ecx,2
     cld
     rep    stosd

    cmp     .dr_fl,15
    jnge    .pee
   @@:
    mov      esi,.sl_ptrs_b_ptr
    mov      edi,[shll]
    cmp      .dr_fl,20
    cmove    esi,.slc_cnt_ptr
    cmove    edi,[cssl]
    cmp      .dr_fl,19
    cmove    esi,.zbuff
    cmove    edi,[sshl]
    cmp      .dr_fl,100
    cmove    esi,[slices_ptrs_buff_ptr]
    mov      .stencil_esi,esi
    mov      .stencil_hor,edi
    mov      esi,[triangles_ptr]
    mov      ecx,[triangles_count_var]
  .again_stencil_tri:
  @@:
    push   ecx
    lodsd
    xchg   eax,ecx
    lodsd
    xchg   ebx,eax
    lodsd
    xchg   eax,ecx
    push   esi
    imul   eax,12
    imul   ebx,12
    imul   ecx,12
    add    eax,.points_rotated_ptr
    add    ebx,.points_rotated_ptr
    add    ecx,.points_rotated_ptr
    movss  xmm3,[eax+8]
    movss  xmm4,[ebx+8]
    movss  xmm5,[ecx+8]
    movups xmm0,[eax]
    movups xmm1,[ebx]
    movups xmm2,[ecx]
    movss  xmm3,[eax+8]
    xorps   xmm4,xmm4
;    movhlps xmm3,xmm0
    movhlps xmm4,xmm1
    movhlps xmm5,xmm2

    shufps  xmm4,xmm4,11110011b

    orps    xmm3,xmm4

    movlhps  xmm3,xmm5	; xmm3 - z float pack
    cvtps2dq xmm0,xmm0
    cvtps2dq xmm1,xmm1
    cvtps2dq xmm2,xmm2
    packssdw xmm0,xmm0
    packssdw xmm1,xmm1
    packssdw xmm2,xmm2
    movd     eax,xmm0
    movd     ebx,xmm1
    movd     ecx,xmm2
    ror      eax,16
    ror      ebx,16
    ror      ecx,16
    movaps   xmm0,xmm3
    mov      edx,.xres_vard
    ror      edx,16
    mov      esi,.stencil_esi
    movd     xmm5,.stencil_hor
    call     stencil_tri
    pop      esi
    pop      ecx
    dec      ecx
    jnz      .again_stencil_tri

    cmp      .dr_fl,20
    jne      .pee

    movzx    eax,.xres_var
    movzx    ebx,.yres_var
    mov      edi,.slc_cnt_ptr
    call     count_transparent_voxels
;    mov      [transparent_voxels_count],ecx
    push     ecx
    shl      ecx,2
    malloc   ecx
    mov      [voxels_depth_ptr],eax

; make background in slices history buff
    cld
    mov      eax,60000.1	  ; clear draw buffer
    mov      edi,[voxels_depth_ptr]
    pop      ecx
    rep      stosd

; fill pointers buff
    mov      eax,[voxels_depth_ptr]
    mov      esi,.slc_cnt_ptr
    mov      edi,.sl_ptrs_b_ptr
    movzx    edx,.xres_var
    movzx    ecx,.yres_var
    imul     ecx,edx
    call     do_variable_slices_ptr_buff  ; file procs_b.inc

   .pee:
    ; or    ebx,ebx
    ; call   peel
     mov    eax,peel	 ; all 4 threaded render modles
     xor    edx,edx
     or     ecx,-1
; in eax  - proc-th adress
;    edx -  param bitewise 'and' with param passed to proc_th
     call   call_thread1

     jmp    .skip_dr

  .no_peel:
    movzx  ebx,.dr_fl
 ;   cmp    ebx,14
 ;   je      .one_th


    add    ebx,ebx
    add    ebx,render_optss + 1

    cmp    byte[ebx],1
    je	   .one_th

  .two_th:
    invoke CreateThread,NULL,NULL,draw_thread,\
	   0,NORMAL_PRIORITY_CLASS,\
	   [ThreadID0]
    mov    [hthread0],eax
 ;   invoke SetThreadPriority,hthread0,NORMAL_PRIORITY_CLASS
    invoke CreateThread,NULL,NULL,draw_thread,\
	   1,NORMAL_PRIORITY_CLASS,\
	   [ThreadID1]
    mov   [hthread1],eax
 ;   invoke SetThreadPriority,hthread1,NORMAL_PRIORITY_CLASS

    invoke SetThreadAffinityMask,[hthread0],0
    invoke SetThreadAffinityMask,[hthread1],0
    invoke ResumeThread,[hthread0]
    invoke ResumeThread,[hthread1]
    invoke WaitForMultipleObjectsEx,2,hthread0,1,-1,0
    invoke CloseHandle,[hthread0]
    invoke CloseHandle,[hthread1]
    jmp    .norm_vect

  .one_th:
    xor   eax,eax
    call  draw_thread

    jmp   .norm_vect

 .skip_dr:
    invoke VirtualFree, [voxels_depth_ptr] , 0,  MEM_RELEASE

  .norm_vect:
    xor    ecx,ecx
    mov    eax,'vert'
    mov    bl,[normals_flag]
    or	   bl,bl
    jz	   .no_nr
    cmp    bl,1
    cmove  eax,ecx
    call   draw_triangles_normals
   .no_nr:


    cmp   [submit_flag],1
    je	  @f
    cmp   [do_submit],1
    jne   .no_submit
    @@:
       mov    esi,.matrix_scaled
       lea    edi,.mx
       call   reverse_mx_3x3
       mov    esi,.points_rotated_ptr
       mov    edi,[points_r_ptr]
       mov    ecx,.points_count_var
       lea    ecx,[ecx*3]
       cld
       rep    movsd

       cvtdq2ps xmm3,[xxadd]
       mov	ebx,[points_r_ptr]
       mov	ecx,.points_count_var
     @@:
       movlps	xmm0,[ebx]
       subps	xmm0,xmm3
       movlps	[ebx],xmm0
       add	ebx,12
       loop	@b

       mov     esi,[points_r_ptr]
       mov     edi,esi ;[edit_particle_no]
       lea     ebx,.mx
       mov     ecx,.points_count_var
       call    rotary
       or      eax,-1
       xor     ebx,ebx
       mov     [edit_particle_no],eax
       mov     dword[edit_end_x],ebx
       mov     dword[edit_start_x],ebx
       mov     [do_submit],bl

       cmp     [submit_flag],1
       jne     @f
       mov     [submit_flag],bl
    ;	movups	xmm0,[NextBendZero]
    ;	movups	xmm1,[NextMsub]
    ;	movlps	xmm2,[NextRotatZero]
    ;	movups	[NextBend1],xmm0
    ;	movups	[NextMxadd],xmm1
    ;	movlps	[NextRotat],xmm2
     lea      esi,[NextMsub]
     lea      edi,[NextMxadd]
     mov      ecx,17
     rep      movsw

       mov	edi,ffd_nodes	      ; zero ffd
       mov	esi,ffd_nodes_intial
       mov	ecx,16*6*3
       cld
       rep	movsd

       call	write_info
       call	NextMdraw_all
       xor	eax,eax
       call	draw_menu

     @@:
  .no_submit:
  ; call       render_fog
  ; call       move_fog_seeds


    cmp        [ffd_flag],0
    jne        @f
    cmp        [bezier_flag],0
    je	       .no_bezA
    cmp        [derive_flag],0
    je	       .no_bezA
  @@:
    mov        edi,[edit_buf_ptr]
    movzx      ecx,.xres_var
    movzx      eax,.yres_var
    imul       ecx,eax
    or	       eax,-1
    cld
    rep        stosd
    mov        esi,ffd_nodes_rot  ;derv_nodes_rot
  ;  mov	eax,ffd_nodes_rot
  ;  cmp	[bezier_flag],0
  ;  cmove	esi,eax
    cmp        [ed_bpatch_derive_on],0
    je	       @f
    movzx      eax,[ed_bpatch_derive_no]
    cmp        ax,0xffff
    je	       @f
    imul       eax,12
    add        esi,eax
    xorps      xmm2,xmm2
    movd       xmm0,[ed_bpatch_derives_x_start]
    movd       xmm1,[ed_bpatch_derives_x_end]
    punpcklwd  xmm0,xmm2
    punpcklwd  xmm1,xmm2
    psubd      xmm1,xmm0
    cvtdq2ps   xmm1,xmm1
    rcpss      xmm2,[scale]
    shufps     xmm2,xmm2,11110000b
    mulps      xmm2,xmm1
    movups     xmm0,[esi]
    addps      xmm0,xmm1
    movlps     [esi],xmm0
  @@:
  .ffd_derv:
    movzx  eax,[bez_patch_count]
    mov    edi,.scr
    mov    ebx,ffd_rect;_d
    mov    ecx,1  ;6
    cmp    [ffd_flag],2
    jne    @f
    mov    ecx,4
  @@:
    cmp    [bezier_flag],0
    cmovne ecx,eax  ;[bez_patch_count]
    cmovne ebx,[derv_rect_ptr]
  @@:
    push   ecx
    push   ebx
    mov    esi,ffd_nodes_rot
    call   draw_bezier_derives1  ; only for sigle patch
    pop    ebx
    add    ebx,16*2
    pop    ecx
    loop   @b

    movzx  eax,[bez_patch_count]
    mov    esi,ffd_nodes_rot
    mov    ebx,ffd_rect
    mov    ecx,1  ;6
    cmp    [ffd_flag],2
    jne    @f
    mov    ecx,4
  @@:
    cmp    [bezier_flag],0
    je	   @f
    mov    ecx,eax  ;[bez_patch_count]
    mov    ebx,[derv_rect_ptr]
  @@:
; in:
;  esi - ptr to derives vertices list
;  ebx - ptr to derives indexes (16 pro one rectangle)
;  ecx - number of surfaces
;  globals - screen_ptr, i12, ..

    call   draw_bezier_bars
  .no_bezA:
	   ; *******************draw edges that intersects
     mov    .tex_ptr,0x0000ff00
     xor    edx,edx
     cmp    [valencEd_flag],1
     jne    .chck_inter
     mov    edx,[edge_s_d_ptr]
   .intersect_valence:
     xor     ecx,ecx
     mov     edi,.edges_ptr
;     mov    .tex_ptr,0x0000ff00
    .checkit:
     push    edx
     push    ecx
     push    edi

     mov     eax,ecx
     mov     ebx,ecx
     shr     eax,3
     and     ebx,111b

     add     eax,edx
     bt      dword[eax],ebx
     jnc     .intersect_loop


     mov       eax,[edi]
     mov       ebx,[edi+4]
     imul      eax,12
     imul      ebx,12
     add       eax,.points_rotated_ptr
     add       ebx,.points_rotated_ptr
     movlps    xmm5,[eax]
     movhps    xmm5,[ebx]
     pxor      xmm6,xmm6
     movd      xmm7, .xres_vard
     punpcklwd xmm7,xmm6
     movlhps   xmm7,xmm7
     movd      edx,xmm7


     cvtps2dq  xmm5,xmm5
     xor       eax,eax


     sub       esp,16
     movups    [esp],xmm5
     pop       eax ebx ecx edx

   ;  movd	xmm3,.col
     movups    xmm5,.width
     mov       edi,plain_horizontal
   ;  call	line_sse
     call      line_grd_tex
   .intersect_loop:
     pop       edi
     add       edi,8
     pop       ecx
     pop       edx
     inc       ecx
   ;  loop	.checkit
     cmp       ecx,[edges_count]
     jnz       .checkit

   .chck_inter:
     mov       ebx,[edges_intersect_ptr]
     xor       eax,eax
     cmp       ebx,eax
     je        .no_inter
     cmp       [mark_coll_ed_flag],al
     je        .no_inter
     cmp       edx,ebx
     je        .no_inter
     mov       edx,ebx
     jmp       .intersect_valence



   .no_inter:
     ; *******************write chunk marks
     xor    eax,eax
     cmp    [show_ch_flag],1
     jne    .no_chunk
     cmp    [chunks_count],eax
     je     .no_chunk
     cmp    [chunks_ptr],eax
     je     .no_chunk

     or     eax,-1	   ;
     mov    edi,[edit_buf_ptr]
     movzx  ebx,.xres_var
     movzx  ecx,.yres_var
     imul   ecx,ebx
     cld
     rep    stosd


     xor       ecx,ecx
     mov       esi,[triangles_ptr]
   .nx_chunk:
     push      ecx
     mov       eax,ecx
     shl       eax,1
     add       eax,[chunks_ptr]
     mov       cx,[eax]
     movzx     ecx,cx
     movd      xmm7,ecx
     ror       cx,2
     mov       edx,ecx
     neg       dx
     ror       dx,3

     mov       [rand_seed],cx
;     mov	ecx,34
;     mov	edx,65535
     call      random
     push      eax
     mov       cx,[rand_seed]
     xchg      cl,ch
 ;    mov	ecx,2
     mov       edx,255
     call      random
     pop       ecx
     shl       ecx,8
     mov       cl,al
     and       ecx,0x00ffffff
     mov       .col,ecx
     cld
     lodsd
     xchg      eax,edi
     lodsd
     xchg      eax,ebx
     lodsd
     xchg      eax,edi

     imul      eax,12
     imul      ebx,12
     imul      edi,12

     push      esi

     add       eax,.points_rotated_ptr
     add       ebx,.points_rotated_ptr
     add       edi,.points_rotated_ptr
     movlps    xmm6,[eax]
     movhps    xmm6,[ebx]
     movlps    xmm0,[edi]
     movhlps   xmm5,xmm6
     addps     xmm0,xmm5
     addps     xmm0,xmm6
     mulps     xmm0,[constrecip3]

     movd      ecx,xmm7
     mov       eax,.col

     movzx     esi,.yres_var
     movzx     edx,.xres_var
     sub       esi,4
     sub       edx,4
      ;  ecx - edge index; signature
      ;  xmm0 - coordinates
      ;  eax - color
      ;  edx  - xresm4
      ;  esi  - yresm4

     call      bar_edge

if 0
     movzx	eax,.yres_var
     movzx	edx,.xres_var
     mov	ecx,4
     sub	eax,ecx ;4
     sub	edx,ecx ;4
  ;   push	 ecx
  ;   push	 ecx
     push	eax
     push	edx
     push	ecx
     push	ecx
     movdqu	xmm5,[esp]
     add	esp,16
     movd	xmm2,edx
     movd	xmm3,.col

     movups	xmm4,[esp]
     pop	eax
     pop	ebx
     pop	ecx
     pop	edx
     movd	xmm1,.scr

     call      linea
end if

   .skip:

     pop       esi
     pop       ecx
     inc       ecx
     cmp       ecx,[triangles_count_var]
     jnz       .nx_chunk
  .no_chunk:



     ; *******************write inner vertices marks
     xor       eax,eax
     cmp       [inner_vert_flag],al
     je        .no_in_vert
     pxor      xmm6,xmm6
     movd      xmm7,.xres_vard
     punpcklwd xmm7,xmm6
     cvtdq2ps  xmm7,xmm7
     movaps    xmm4,[const4]
     subps     xmm7,xmm4
     mov       edi,0x000fffff	 ; col
     movd      xmm6,edi
     pshufd    xmm6,xmm6,0
     movzx     edx,.xres_var
     shl       edx,2
     mov       ecx,.points_count_var
     mov       esi,.inner_vert_ptr
    .nx_in_vert:
     push      ecx
     mov       edi,ecx
     mov       eax,ecx
     shr       eax,3
     and       ecx,111b
     add       eax,esi
     bt        [eax],ecx   ; if bit is set vertex is not inside
     jc        .end_in_vrtl

     imul      edi,12
     add       edi,.points_rotated_ptr
     movlps    xmm0,[edi]
     movaps    xmm1,xmm0
     movaps    xmm2,xmm0
     cmpltps   xmm1,xmm4 ;[const4]
     cmpltps   xmm2,xmm7
     pxor      xmm2,xmm1
     movd      eax,xmm2
     or        eax,eax
     jz        .end_in_vrtl

     cvtps2dq  xmm0,xmm0
     movd      eax,xmm0
     psrldq    xmm0,4
     movd      ebx,xmm0
     dec       eax
     dec       ebx
     shl       eax,2
     imul      ebx,edx
     add       ebx,eax
     add       ebx,.scr
     movq      [ebx],xmm6
     movd      [ebx+8],xmm6
     add       ebx,edx
     movq      [ebx],xmm6
     movd      [ebx+8],xmm6
     add       ebx,edx
     movq      [ebx],xmm6
     movd      [ebx+8],xmm6
   .end_in_vrtl:

     pop	ecx
     loop	.nx_in_vert

  .no_in_vert:

    ; ******************triangulize area/rect
    cmp      [set_tri_area],0
    jne      @f 		      ;
    call     draw_triangulize_rect
  @@:
    call     write_info
    call     NextMdraw_all
    stretch_menu

    add      esp,128
    pop      ebp

    pop  edi esi ebx
ret


;==================================================================
;==================================================================
;==================================================================
;==================END MAIN RENDERING PROC=========================
;==================================================================
;==================================================================
;==================================================================

normalize_all_light_vectors:
	mov	edi,lights
     @@:
	call	normalize_vector	   ;	   3dmath.inc
	add	edi,LIGHT_SIZE
	cmp	edi,lightsend
	jl	@b
ret

read_hrt:
     ; partly generated 3d object
     push	ebp
     mov	ebp,esp
     sub	esp,16

     .val	equ dword[ebp-4]
     .val2	equ dword[ebp-8]
     .val3	equ dword[ebp-12]
     .val4	equ dword[ebp-16]

   ..factor  = 2

    mov    esi,long_pipe_derv_init
    mov    edi,long_pipe_derv+12
    mov    ecx,32 * ..factor
    lea    ecx,[ecx*3 + 8]
    or	   edx,-1
    call   init_long_pipe_derives
    push   dword long_pipe_derv
    push   dword 16 * 16 * ..factor  + 8
    call   normalize_object1


       mov     edx,[long_pipe_vert_ptr]
       mov     eax,(2 + 3) shl 2	   ; one segment
       mov     ebx,32 * 3 * ..factor + 8   ; segments count
       mov     ecx,long_pipe_derv
       mov     esi,1  ; bspl [lpcurve_tp]
       call    calc_long_curve
       call    zero_flags
       ; memory work
       xor     eax,'frea'
       call    free_mem_for_tp1
       mov     [triangles_count_var],50000
       mov     [points_count_var],95500
       mov     eax,'alla'
       call    alloc_mem_for_tp1

       mov    esi,[long_pipe_vert_ptr]
       mov    ecx,32 * 3 * ..factor * (2 + 3 shl 2) ; whole pipe verts count
       ; calc long pipe
       mov    eax,8	  ; rotary count per one turn
       mov    edx,0.019   ; radius
       movd   xmm0,edx
       mov    ebx,[triangles_ptr]
       mov    edi,[points_r_ptr]
; in:  esi  - ptr to curve vertices - middle points of pipe
;      ecx  - curve vertices count
;      eax  - rotary steps count as integer
;      xmm0 - radius, lowest dword float
;      ebx  - ptr to triangles list
;      edi  - ptr to p ipe vertices
; out: ecx  - triangles count
;      ebx  - vertices count

       call   do_long_pipe
       mov    [triangles_count_var],ecx
       mov    [points_count_var],ebx
       xchg   ecx,ebx

       mov    edi,[points_r_ptr]
     @@:
       mov    eax,[edi+8]
       mov    edx,[edi+4]

       mov    [edi+4],eax
       mov    [edi+8],edx
       add    edi,12
       loop   @b

	push   edi  ; end
	push   [points_r_ptr]
	push   ebx
	call   normalize_object1

	mov	.val,1.43
	mov	.val2,0.12
	mov	.val3,0.6

	mov	edx,[points_r_ptr]
	mov	ecx,[points_count_var]
	fninit
      .fix:
	fld	dword[edx]
	fld	st
	fmul	.val
	fstp	dword[edx]
	fld	dword[edx+4]
	fadd	.val2
	fstp	dword[edx+4]
	fcos
	fmul	.val3
	fmul	dword[edx+8]
	fstp	dword[edx+8]
       @@:
	add	edx,12
	loop	.fix


	mov    edx,[triangles_ptr]	 ; sense 'z' coof of nr vec fix
	mov    ecx,[triangles_count_var]
      @@:

	mov    eax,[edx]
	push   dword[edx+4]
	pop    dword[edx]
	mov    [edx+4],eax
	add    edx,12
	loop   @b



	pop    edi
	push   [points_count_var]
	mov    esi,hrt_points
	mov    ecx,hrt_points_count
	push   edi   ; first param of norm_ob
	push   ecx   ; sec param
	pxor   xmm0,xmm0

      @@:
	pxor	  xmm2,xmm2
	movq	  xmm1,[esi]
	punpcklwd xmm1,xmm0
	pcmpgtw   xmm2,xmm1
	pslldq	  xmm2,2
	por	  xmm1,xmm2
	cvtdq2ps  xmm1,xmm1
	movups	  [edi],xmm1
	inc	  [points_count_var]
	add	  edi,12
	add	  esi,6
	loop	  @b

	call	  normalize_object1




	mov    edi,[triangles_ptr]
	mov    ebx,[triangles_count_var]
	imul   ebx,12
	add    edi,ebx
	pop    edx
	mov    esi,hrt_triangles
	mov    ecx,hrt_triangles_count
	add    [triangles_count_var],ecx
	cld
	lea    ecx,[ecx*3]
     @@:
	xor    eax,eax
	lodsw
	add    eax,edx
	stosd
	loop   @b

	mov    esp,ebp
	pop    ebp

ret

alloc_mem_for_tp1:

;; in eax=1 - free all mem , full alloc
;;    eax=0 - not free mem, alloc all but [triangles_ptr], and [points_r_ptr]
;;    eax = 100 - alloc only for triangles_ptr and points_r_ptr not free at all
;;    eax=2 - not free mem full alloc


if 0 ;abandon text below
;; and in this cause (eax<>1) - in: eax -> points (vertices) count
;;				    ebx -> triangles count
;;				out: eax -> point (vertices) ptr
;;				     ebx -> triangles ptr
end if ; abandon text above
;     cmp     eax,3
;     je      @f
     push     ebp
     mov      ebp,esp

     push     eax
     push     dword 12
     mov      edx,[points_count_var]
     mov      ebx,[triangles_count_var]
     mov      ecx,30000
     cmp      edx,ecx	       ; set minimum
     cmovb    edx,ecx
     cmp      ebx,ecx
     cmovb    ebx,ecx
     push     edx
     push     ebx
   ;  push     [points_count_var]
   ;  push     [triangles_count_var]


     .mark   equ dword[ebp-4]
     .i12    equ dword[ebp-8]
     .pcv    equ dword[ebp-12]
     .tcv    equ dword[ebp-16]

     cmp     eax,'alla'      ; whole allocation
     je      .l1
     cmp     eax,'tpon'      ; tpon only for tri and vert
     je      .l1
     or      eax,eax
     jz      .no_tp
     cmp     eax,2
     je      .l1
     cmp     eax,100
     je      .l1
     call    free_mem_for_tp1
;     jmp     .no_tp
  .l1:
     mov     eax,.pcv
     imul    eax,12
     add     eax,40
     malloc  eax

     mov     [points_r_ptr],eax

     mov     eax,.tcv
     imul    eax,12
     add     eax,65536
     malloc  eax

     mov    [triangles_ptr],eax
     cmp    .mark,'tpon'
     je     .end
     cmp    .mark,100
     je     .end

 .no_tp:
 .n_ch:
     mov     eax,.tcv
     shl     eax,3
     lea     eax,[eax*3]
     add     eax,65536
     malloc  eax
     mov     [edges_ptr],eax

   ;  mov     eax,.pcv
   ;  imul    eax,12
   ;  add     eax,20
   ;  malloc  eax
   ;  mov     [shadow_vertices_ptr],eax

     mov     eax,.tcv
     imul    eax,12
     add     eax,65536
     malloc  eax
     mov     [triangles_normals_ptr],eax

     mov     eax,.tcv
     imul    eax,12
     add     eax,65536
     malloc  eax
     mov     [triangles_normals_rotated_ptr],eax

     mov     ebx,.pcv
     imul    ebx,12
     add     ebx,65536
     malloc  ebx
     mov     [points_normals_ptr],eax

     malloc ebx
     mov    [shadow_vertices_ptr],eax

     malloc  ebx
     mov     [points_normals_rotated_ptr],eax

     malloc  ebx
     mov     [points_rotated_ptr],eax

     mov     ebx,.pcv
     shl     ebx,2
     add     ebx,20
     malloc  eax
     mov     [tex_points_ptr],eax

     shl     ebx,1
     malloc  ebx
     mov     [tex_points_f_ptr],eax

   .end:
     mov     esp,ebp
     pop     ebp
ret

free_mem_for_tp1:

;    eax = 33 -> not free for tris and verts
   push    ebp
   mov	   ebp,esp

   mov	   ebx,MEM_RELEASE

   cmp	   eax,'frea' ; free all
   je	   @f
   cmp	   eax,33
   je	   .no_tp
  @@:
   invoke  VirtualFree,[points_r_ptr] , 0,ebx
   invoke  VirtualFree,[triangles_ptr] , 0, ebx
 .no_tp:
   invoke  VirtualFree,[edges_ptr] , 0, ebx
   invoke  VirtualFree,[shadow_vertices_ptr] , 0, ebx
   invoke  VirtualFree,[triangles_normals_ptr] , 0, ebx
   invoke  VirtualFree,[triangles_normals_rotated_ptr]	, 0, ebx
   invoke  VirtualFree,[points_normals_ptr] , 0,  MEM_RELEASE
   invoke  VirtualFree,[points_normals_rotated_ptr] , 0, ebx
   invoke  VirtualFree,[points_rotated_ptr] , 0,  ebx
   invoke  VirtualFree,[tex_points_ptr], 0, ebx
   invoke  VirtualFree,[tex_points_f_ptr] , 0, ebx
   invoke  VirtualFree,[chunks_ptr] , 0, ebx
   mov	   esp,ebp
   pop	   ebp
ret

read_tp_variables1:	       ; read [triangles_count_var] and ...
			       ; ..[points_count_var]
; in:	eax -> ptr to 3ds file
;	ebx -> file size
;	edx = 1 -> read chunks count
; out:	ecx -> points count
;	ebx -> triangles count
;	edx -> chunks number
	push ebp
	mov  ebp,esp
	sub  esp,30
	.tri_count  equ [ebp-4]
	.vert_count equ [ebp-8]
	.FileSize   equ [ebp-12]
	.EndFile    equ [ebp-16]
	.fe	    equ [ebp-16]
	.chunks_c   equ dword[ebp-20]
	.rch	    equ dword[ebp-24]
	.fem2	    equ dword[ebp-28]
	.fem1	    equ dword[ebp-32]

	xor	 esi,esi
	mov	.rch,edx
	mov	.chunks_c,esi
	mov	.FileSize,ebx
	xor	ebx,ebx
	mov	.vert_count,ebx
	mov	.tri_count,ebx
	mov	esi,eax  ;[File_ptr]
  ;	 int3

	cmp	[esi],word 4D4Dh
	je	@f ;Must be legal .3DS file
	xor	eax,eax
	ret
    @@:
	mov	eax,dword[esi+2]
	cmp	eax,.FileSize ;This must tell the length
	je	@f
	xor	eax,eax
	ret
     @@:
	add	eax,esi
	mov	.EndFile,eax
	dec	eax
	mov	.fem1,eax
	dec	eax
	mov	.fem2,eax
	xor	eax,eax
	add	esi,6
      @@:
	cmp	[esi],word 3D3Dh
	je	@f

	cmp	esi,.fem2 ;.EndFile
	ja	.ex1

	add	esi,[esi+2]
	jmp	@b
      @@:
	add	esi,6
    ;  .find4k:

      .find4k:
	cmp	[esi],word 4000h
	je	@f
	cmp	esi,.fem2
	ja     .ex1

	add	esi,[esi+2]
	cmp	esi,.fe  ;.FileSize
	jc	.find4k
	jmp	.exit
      @@:
	add	esi,6
      @@:
	cmp	esi,.fe
	ja	.ex1
	cmp	[esi],byte 0
	je	@f
	inc	esi
	jmp	@b
      @@:
	inc	esi
      @@:
	cmp	esi,.fe ;m2
	ja	.ex1
	cmp	[esi],word 4100h
	je	@f
	cmp	esi,.fem2
	ja     .ex1	     ; some 3ds files needs thise check

	add	esi,[esi+2]
	jmp	@b
      @@:
	add	esi,6
      @@:
	cmp	esi,.fe ;m2
	ja	.ex1
	cmp	[esi],word 4110h
	je	@f

	cmp	esi,.fem2
	ja	.ex1




	cmp	dword[esi+2],eax
	je	.find4k

	add	esi,[esi+2]
	jmp	@b
      @@:
	movzx	ecx,word[esi+6]
	add	.vert_count,ecx
	mov	edx,ecx
	add	esi,8
	lea	ecx,[ecx*3]
	add	ecx,ecx
	add	ebx,ecx
	lea	esi,[esi+ecx+ecx]
      @@:
	cmp	[esi],word 4120h
	je	@f
	cmp	esi,.fem2	 ; some 3ds files needs thise check
	ja	.ex1
	add	esi,[esi+2]
	jmp	@b
      @@:
	movzx	ecx,word[esi+6]
	inc	.chunks_c
	add	.tri_count,ecx
	add	esi,8
     ;	 int3
	shl	ecx,3
	add	esi,ecx
     ;	 add	 eax,edx
	jmp	.find4k
    ;	 mov	 eax,-1 ;<---mark if OK
     .ex1:
     .exit:
	mov	ecx,dword .vert_count
	mov	ebx,dword .tri_count
    ;	 int3
	xor	edx,edx
	cmp	.rch,1
	jne	@f
	mov	edx,.chunks_c
  ;    .ex1:
      @@:
	mov	esp,ebp
	pop	ebp
ret

;free_screen_mem:
;  not used
; params - some globals
;     invoke   VirtualFree ,[edit_tri_area_ptr],0 , MEM_RELEASE
;     invoke   VirtualFree ,[slices_counter_ptr],0 , MEM_RELEASE
;     invoke   VirtualFree ,[screen_ptr],0 , MEM_RELEASE
;     invoke   VirtualFree ,[Zbuffer_ptr],0 , MEM_RELEASE
;     invoke   VirtualFree ,[edit_buf_ptr],0 , MEM_RELEASE
;     invoke   VirtualFree ,[shd_buffer_ptr],0 , MEM_RELEASE
;
;ret

; read from 3ds file
;in:  eax - ptr to triangles
;     ebx - ptr to vertices
;     ecx - ptr to file begin
;     edx - ptr to file end
;out: ecx - points (vertices) count
;     edx - triangles count

read_from_file:
;; this procedure coded orginally by Madis Kalme
;; Thanks a lot Madis!!!
	push	     ebp
	mov	     ebp,esp
	sub	     esp,44
	.triangles_p equ [ebp-4]
	.vert_p      equ [ebp-8]
	.file	     equ [ebp-12]
	.file_end    equ [ebp-16]
	.fe	     equ [ebp-16]    ; alias
	.vert_cnt    equ [ebp-20]
	.tri_cnt     equ [ebp-24]
	.cur_ch      equ dword[ebp-28]
	.cur_chp     equ dword[ebp-32]
	.cccnt	     equ dword[ebp-36]
	.fem2	     equ dword[ebp-40]
	.fem1	     equ dword[ebp-44]

	mov	.triangles_p,eax
	mov	.vert_p,ebx
	mov	.file,ecx
	mov	edi,eax  ;[triangles_ptr]
	xor	ebx,ebx
	mov	dword .vert_cnt,0
	mov	dword .tri_cnt,0
	mov	esi,ecx
	cmp	[esi],word 4D4Dh
	jne	.exit ;Must be legal .3DS file
;	 cmp	 dword[esi+2],EndFile-SourceFile
;	 jne	 .exit ;This must tell the length
	mov	eax,dword[esi+2]
  ;	 cmp	 eax,[fsize]
  ;	 jne	 .exit
	add	eax,esi
	mov	.file_end,eax
	dec	eax
	mov	.fem1,eax
	dec	eax
	mov	.fem2,eax
	xor	eax,eax
	mov	.cccnt,eax
	add	esi,6
      @@:
	cmp	[esi],word 3D3Dh
	je	@f
	cmp	esi,.fem2	; some 3ds files needs thise check
	ja	.ex1

	add	esi,[esi+2]
	jmp	@b
      @@:
	add	esi,6
      .find4k:
	cmp	[esi],word 4000h
	je	@f
	cmp	esi,.fem2
	ja     .ex1

	add	esi,[esi+2]
	cmp	esi,.file_end
	jc	.find4k
	jmp	.exit
      @@:
	add	esi,6
      @@:
	cmp	[esi],byte 0
	je	@f
	cmp	esi,.fem1
	ja     .ex1
	inc	esi
	jmp	@b
      @@:
	inc	esi
      @@:
	cmp	esi,.fe
	ja	.ex1
	cmp	[esi],word 4100h
	je	@f
	cmp	esi,.fem2
	ja     .ex1	     ; some 3ds files needs thise check

	add	esi,[esi+2]
	jmp	@b
      @@:
	add	esi,6
      @@:
	cmp	esi,.fe  ;m2
	ja	.ex1
	cmp	[esi],word 4110h
	je	@f

	cmp	dword[esi+2],0
	je	.find4k

	add	esi,[esi+2]
	jmp	@b
      @@:
	movzx	 ecx,word[esi+6]       ; word
	add	 .vert_cnt,ecx
	mov	 edx,ecx
	add	 esi,8
	push	 edi
	mov	 edi,.vert_p
	or	 ecx,ecx
	jz	 .done
     @@:
	cld
	lea	 ecx,[ecx*3]
	rep	 movsd
	mov	 .vert_p,edi
     .done:
	pop	 edi
      @@:
	cmp	[esi],word 4120h
	je	@f
	add	esi,[esi+2]
	jmp	@b
      @@:
	movzx	ecx,word[esi+6]
	add	.tri_cnt,ecx
	add	esi,8
	movd	xmm1,.cccnt  ;eax
	pshufd	xmm1,xmm1,0
	movd	xmm5,.cur_ch
	pshuflw xmm5,xmm5,01000100b
	pshufd	xmm5,xmm5,0
     @@:
     .two:
	pxor	  xmm3,xmm3
	movq	  xmm0,[esi]
	punpcklwd xmm0,xmm3 ;[the_zero]
	paddd	  xmm0,xmm1
	movq	  [edi],xmm0
	pshufd	  xmm0,xmm0,00000010b
	movd	  [edi+8],xmm0
	add	  edi,12
	add	  esi,8

       @@:
	loop	  .two
	inc	  .cur_ch
	add	  .cccnt,edx
	jmp	  .find4k
      .ex1:
	or	   eax,-1 ;<---mark if OK
      .exit:
	mov	   ecx,dword .vert_cnt
	mov	   edx,dword .tri_cnt
	mov	   ebx,.file_end
	mov	   esp,ebp
	pop	   ebp
ret

;====================================================================
init_triangles_normals: ; orginally by Mikolaj Feliks
			 ; some changes by macgub
	push	 ebp
	mov	 ebp,esp
	sub	 esp,20

	.pts_ptr equ dword[ebp-4]
	.vec	 equ dword[ebp-20]
	prompt	 prompt_norm

	push	 [points_r_ptr]
	pop	 .pts_ptr

	mov	 ecx,[triangles_count_var]
	mov	 edi,[triangles_normals_ptr]
	mov	 esi,[triangles_ptr]

	cld
     @@:
	lodsd
	xchg	 eax,ebx
	lodsd
	xchg	 edx,eax
	lodsd

	imul	 eax,12
	imul	 ebx,12
	imul	 edx,12
	add	 eax,.pts_ptr
	add	 ebx,.pts_ptr
	add	 edx,.pts_ptr
	push	 edi
	movups	 xmm0,[eax]
	movups	 xmm1,[ebx]
	movups	 xmm2,[edx]
	subps	 xmm0,xmm1
	subps	 xmm1,xmm2
	call	 cross_reg
	lea	 edi,.vec
	movups	 [edi],xmm0
	call	 normalize_vector
	movups	 xmm0,[edi]
	pop	 edi
;	 movups
;	 movaps   xmm2,xmm0
 ;	 mulps	  xmm0,xmm0
 ;	 movhlps  xmm1,xmm0
 ;	 haddps   xmm0,xmm0
 ;	 addps	  xmm1,xmm0
 ;	 shufps   xmm1,xmm1,0
 ;	 mulps	  xmm2,xmm0

	movups	 [edi],xmm0
	add	 edi,12
	loop	 @b


	cls
	mov	esp,ebp
	pop	ebp
ret
;=====================================================================
init_pt_norm_th:
;    if edx =  3 then esi = .tri_ch
;		      edi = .t_ptr

    .tri_ch   equ dword[ebp-4]
    .t_ptr    equ dword[ebp-8]
    .val      equ dword[ebp-8]

    push     ebp
    mov      ebp,esp
    push     esi
    push     edi
    push     eax


    push     edx
    cmp      edx,3
    jne      @f
    mov      edx,0x80000000
    mov      [thread_params],esi
    mov      [thread_params+4],edi
   @@:
    cmp      edx,1
    jne      @f
    xor      edx,edx
  @@:
    or	     ecx,-1
    mov      eax,init_point_normals
    call     call_thread1
    pop      edx
    cmp      edx,3
    jne      @f

    invoke   VirtualFree ,.t_ptr,0 , MEM_RELEASE
    invoke   VirtualFree ,.tri_ch,0 , MEM_RELEASE

   @@:


    mov      esp,ebp
    pop      ebp
ret

;=================================================================================

init_point_normals:

.z	      equ  dword [ebp-8]
.y	      equ  dword [ebp-12]
.x	      equ  [ebp-16]
.point_number equ  [ebp-32]
.hit_faces    equ  dword [ebp-36]
.end_point    equ  dword [ebp-40]
.chun	      equ  dword [ebp-44]
.tri_ch       equ  dword [thread_params]
.t_ptr	      equ  dword [thread_params+4]
.max_val      equ  dword [ebp-48]     ; max val of .tri_ch  list
.triangles_count_var equ dword[ebp-52]


;    if most sig bit of ebx is set then  [thread_params] = .tri_ch
;					 [thread_params+4] = .t_ptr

	push	  ebp
	mov	  ebp,esp
	sub	  esp,100
	and	  ebp,0xfffffff0
	prompt	  prompt_norm	; macro - draw info
	mov	  eax,ebx
	shr	  eax,31

	mov	  .chun,eax
	push	  [triangles_count_var]
	pop	  .triangles_count_var

	and	  ebx,0xffff
	cmp	  ebx,100
	jne	  @f
	xor	  edx,edx
	xor	  esi,esi
	push	  [points_count_var]
	pop	  .end_point
	jmp	  .do
      @@:
	mov	  edx,[points_count_var]
	mov	  edi,[points_count_var]
	and	  edi,011b
	shr	  edx,2
	mov	  ecx,edx
	imul	  edx,ebx    ; edx  - first vert no
	add	  ecx,edx    ; ecx  - end vert no

	cmp	  ebx,3
	jne	  @f
	add	  ecx,edi
      @@:
	mov	  .end_point,ecx

	cmp	  .chun,1
	jne	  .do

	mov	  ecx,.triangles_count_var
	shl	  ecx,3
	lea	  ecx,[ecx*3]
	add	  ecx,.tri_ch
	mov	  .max_val,ecx

     .lp1:
	mov	  ebx,edx
	shl	  ebx,2
	add	  ebx,.t_ptr
	mov	  esi,[ebx]
	or	  esi,esi
	jz	  .old

	xorps	  xmm1,xmm1
	xor	  ecx,ecx
     @@:
	mov	  eax,[esi+4] ; eax - tri index
	mov	  ebx,[esi]
	imul	  eax,12
	add	  eax,[triangles_normals_ptr]
	movups	  xmm0,[eax]
	inc	  ecx
	addps	  xmm1,xmm0
	add	  esi,8
	cmp	  esi,.max_val	 ; some objects need this check
	ja	  .old		 ;old method
	cmp	  ebx,[esi]
	je	  @b

	cvtsi2ss xmm2,ecx
	rcpss	 xmm2,xmm2
	shufps	 xmm2,xmm2,0
	mulps	 xmm1,xmm2
	mov	 edi,edx
	imul	 edi,12
	add	 edi,[points_normals_ptr]
	movlps	 [edi],xmm1
	movhlps  xmm1,xmm1
	movss	 [edi+8],xmm1
	call	  normalize_vector

	inc	 edx
	cmp	 edx,.end_point
	jnz	 .lp1

	jmp	 .end


     .old:

     .do:
       mov    dword .point_number,edx

    .ipn_loop:
	xor	  esi,esi
	xorps	  xmm3,xmm3
	mov	  .hit_faces,esi
	movups	  .x,xmm0
	mov	  esi,[triangles_ptr]
	mov	  ecx,[triangles_count_var]
	lea	  ecx,[ecx*3]

    .ipn_check_vertices:
      @@:
      .check_one_vertex:
	mov	  eax,dword[esi]    ;  eax - point_number
	cmp	  eax,.point_number
	je	  .inc_h_f ;   .ipn_next_vertex
      .inc_counters:
	add	  esi,4
	loop	  .calc_one_normal
	dec	  ebx
	jnz	  .check_one_vertex
	jmp	  .ipn_check_vertices
      .inc_h_f:
	push	  eax
	push	  edx

	mov	  eax,esi
	sub	  eax,[triangles_ptr]
	add	  eax,[triangles_normals_ptr]

	movups	  xmm0,[eax]
	addps	  xmm3,xmm0
	addps	  xmm0,.x
	movaps	  .x,xmm0

	pop	 edx
	pop	 eax
	inc	 .hit_faces
	jmp	 .inc_counters

     .calc_one_normal:
	mov	  edi,.point_number
	imul	  edi,12
	add	  edi,[points_normals_ptr]


	cvtsi2ss   xmm1,.hit_faces
	rcpss	   xmm1,xmm1
	shufps	   xmm1,xmm1,0
	mulps	   xmm3,xmm1
	movhlps    xmm2,xmm0
	movlps	   [edi],xmm0
	movss	   [edi+8],xmm2

	call	  normalize_vector
	inc	  dword .point_number
	mov	  edx,.point_number	 ;xmm0
	cmp	  edx,.end_point
	jne	  .ipn_loop
     .end:

	cls	  ; macro
	add	  esp,100
	pop	  ebp

ret
;=============================================================

draw_menu:
;    in:
;	eax <> 0 - skip clear monu screen buffer
    or	   eax,eax
    jnz    .ffa

    mov    edi,menu_screen
    mov    ecx,MXRES*MYRES   ;+ IN_CNST)
    xor    eax,eax
    cld
    rep    stosd

  .ffa:



    mov    esi,menu_data
  .lab:
    cld
    push   esi
    mov    ax,[esi+2]
    shl    eax,16
    mov    ax,[esi+6]
    mov    bx,[esi+4]
    shl    ebx,16
    mov    bx,[esi+8]
    push   esi
    mov    esi,MXRES
    movd   xmm2,esi
    movd   xmm7,esi
    mov    esi,menu_screen
    movd   xmm1,esi
    mov    esi,0x0000ff00
    movd   xmm3,esi
    mov    esi,MYRES
    movd   xmm5,esi
    pshufd xmm5,xmm5,10010011b
    por    xmm7,xmm5
    pxor   xmm6,xmm6
    call   write_frame
    pop    esi
    mov    eax,[esi+2]	    ; write button desc
    ror    eax,16
    add    eax,0x00030000
    mov    ebx,font_table + 62 ; /   font_tab2 + 62
    mov    ecx,MXRES shl 16 + MYRES
    add    esi,10
    mov    edi,menu_screen
    mov    edx,16 ; / 9
    call   write_text1
    add    eax,90 shl 16 + 0	  ; write flag of button
    or	   ebx,-1
    cmp    dword[esi+13],ebx
    je	   @f
    movzx  edi,byte[esi+12]
    imul   edi,5  ;[i5]
    mov    ebx,font_table + 62 ; /   font_tab2 + 62
    mov    esi,[esi+13]
    add    esi,edi
    mov    edi,menu_screen
    mov    edx,16 ; / 9
    call   write_text1
   @@:
    pop    esi
    add    esi, 33
    or	   ebx,-1
    cmp    word[esi],bx
    jne    .lab

    stretch_menu

ret
;=================================================================
random_point_light_position:

;  rdtsc
;  mov	    [rand_seed],ax
  xor	   ebx,ebx
 @@:
  push	   ebx
  mov	   ecx, -1000
  mov	   edx, 1000
  call	   random
  cvtsi2ss xmm0,eax
  pop	   ebx
  movss    [ppoint_light_coords+ebx],xmm0
  add	   ebx,4
  cmp	   ebx,8
  jnz	   @b

  xor	   ebx,ebx
  mov	   eax,255
  cvtsi2ss xmm1,eax   ;[i255d]	;eax
 @@:
  push	   ebx
  xor	   ecx,ecx
  mov	   edx,255
  call	   random
  cvtsi2ss xmm0,eax
  divss    xmm0,xmm1
  pop	   ebx
  movss    [plane_equation+ebx],xmm0
  add	   ebx,4
  cmp	   ebx,8
  jnz	   @b
ret
 ;===========================================================

mem_error:
  invoke  MessageBox,NULL,_error_mem,NULL,MB_ICONERROR+MB_OK
  invoke  Sleep,500
  invoke  ReleaseDC,[hwnd],[hdc]
  invoke  DestroyWindow,[hwnd]
  invoke  ExitProcess,[msg.wParam]
ret

include 'tesselate.inc'
include '3tex_cat.inc'
include '3flat_cat_floatz.inc'
include '3dmath.inc'
include 'procs_a.inc'
include 'procs_b.inc'
include 'procs_c.inc'
include 'chunks.inc'
include '3bump_cat.inc'
include '3bump_tex.inc'
include '3two_tex.inc'
include '3grd_tex.inc'
include '3grd_cat.inc'
include '3r_phg_col_stencil.inc'
include '3bi_fil.inc'
include 'grd_line.inc'
include 'line.inc'
include '3shd.inc'
include 'arbitrary_vector.inc'
include 'asc.inc'
include 'asc_header.inc'
include 'menu.inc'
include 'edit.inc'
include 'draw_thrd.inc'
include 'bezier_surface.inc'
include '3stencil.inc'
include '3glass.inc'
include '3glass_tex.inc'
include 'long_pipe.inc'
include '3glass_tex_slices.inc'
include '3slices_counter.inc'
include '3stencil_slices.inc'
include '3stencil_slices_var.inc'
include '3glass_tex_slices_var.inc'
include 'normals.inc'
include '3fur.inc'
include '3ray_shd.inc'
include 'file.inc'
include 'sort.inc'
;include 'calc_cones.inc'  cones disabled
 include '3normal.inc'

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; data
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
section '.data' data readable writeable
   _class	TCHAR	   'Win3ds',0
   _title	TCHAR	   'Ave Cruce Salus Mea',0
   _error	TCHAR	   'Startup failed.',0
   _error_mem	TCHAR	   'Memory fail.',0
   fakebitmap:	dd 10 dup     20
   include     'data.inc'

section     '.idata' import data readable writeable
    library kernel32,'KERNEL32',user32,'USER32',gdi32,'GDI32',winmm,'WINMM', comdlg32,'COMDLG32'    ;	timing,'timing.dll'
    include '.\include\api\kernel32.inc'
    include '.\include\api\user32.inc'
    include '.\include\api\gdi32.inc'
    include '.\include\api\comdlg32.inc'

